from django.urls import path

from ddm.auth import views


app_name = 'ddm_auth'
urlpatterns = [
    path(r'projects/<slug:project_url_id>/token/', views.ProjectTokenView.as_view(), name='project_token'),
    path(r'auth/no-permission/', views.DdmNoPermissionView.as_view(), name='no_permission'),
]



from django.test import SimpleTestCase
from django.urls import reverse, resolve
import ddm.participation.views as participation_views


class TestParticipationFlowUrls(SimpleTestCase):

    def test_project_briefing_url_resolves(self):
        url = reverse('ddm_participation:briefing', args=['project-slug'])
        self.assertEqual(
            resolve(url).func.view_class,
            participation_views.BriefingView
        )

    def test_project_datadonation_url_resolves(self):
        url = reverse('ddm_participation:datadonation', args=['project-slug'])
        self.assertEqual(
            resolve(url).func.view_class,
            participation_views.DataDonationView
        )

    def test_project_questionnaire_url_resolves(self):
        url = reverse('ddm_participation:questionnaire', args=['project-slug'])
        self.assertEqual(
            resolve(url).func.view_class,
            participation_views.QuestionnaireView
        )

    def test_project_debriefing_url_resolves(self):
        url = reverse('ddm_participation:debriefing', args=['project-slug'])
        self.assertEqual(
            resolve(url).func.view_class,
            participation_views.DebriefingView
        )



from django.urls import path, include
from django.views.generic import RedirectView

urlpatterns = [
    path(r'', RedirectView.as_view(pattern_name='ddm_projects:list'), name='ddm_landing'),
    path(r'studies/<slug:slug>/',
         include('ddm.participation.urls', namespace='ddm_participation')),
    path(r'projects/',
         include('ddm.projects.urls', namespace='ddm_projects')),
    path(r'projects/<slug:project_url_id>/questionnaire/',
         include('ddm.questionnaire.urls', namespace='ddm_questionnaire')),
    path(r'projects/<slug:project_url_id>/data-donation/',
         include('ddm.datadonation.urls', namespace='ddm_datadonation')),
    path(r'', include('ddm.logging.urls', namespace='ddm_logging')),
    path(r'', include('ddm.auth.urls', namespace='ddm_auth')),
    path(r'api/', include('ddm.apis.urls', namespace='ddm_apis')),
]



import random
import string

from django.db import migrations, models

def generate_unique_url_id(existing_ids):
    """Generate a unique external_id."""
    while True:
        url_id = ''.join(random.SystemRandom().choice(string.ascii_letters + string.digits) for _ in range(8))
        if url_id not in existing_ids:  # Ensure uniqueness
            return url_id

def populate_url_id_of_existing_projects(apps, schema_editor):
    DonationProject = apps.get_model('ddm_projects', 'donationproject')
    existing_ids = set(DonationProject.objects.values_list('url_id', flat=True))

    objects_to_update = []
    for project in DonationProject.objects.all():
        if project.url_id is None:
            project.url_id = generate_unique_url_id(existing_ids)
            existing_ids.add(project.url_id)
            objects_to_update.append(project)

    DonationProject.objects.bulk_update(objects_to_update, ['url_id'])


class Migration(migrations.Migration):

    dependencies = [
        ('ddm_projects', '0003_donationproject_active_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='donationproject',
            name='url_id',
            field=models.CharField(max_length=8, unique=True, null=True),
            preserve_default=False,
        ),
        migrations.RunPython(populate_url_id_of_existing_projects),
        migrations.AlterField(
            model_name='donationproject',
            name='url_id',
            field=models.CharField(max_length=8, unique=True, null=False),
        ),
    ]



import datetime

from django.test import TestCase
from django.utils import timezone
from django.contrib.auth import get_user_model
from rest_framework import exceptions

from ddm.auth.models import ProjectTokenAuthenticator, ProjectAccessToken
from ddm.projects.models import ResearchProfile, DonationProject


User = get_user_model()


class TestCustomTokenAuthenticator(TestCase):

    @classmethod
    def setUpTestData(cls):
        # User
        base_creds = {
            'username': 'base_user', 'password': '123', 'email': 'base@mail.com'
        }
        base_user = User.objects.create_user(**base_creds)
        base_user_profile = ResearchProfile.objects.create(user=base_user)

        # Project
        cls.project = DonationProject.objects.create(
            name='Base Project', slug='base', owner=base_user_profile
        )

        # Authenticator
        cls.authenticator = ProjectTokenAuthenticator()
        cls.token = ProjectAccessToken.objects.create(
            project=cls.project, created=timezone.now(), expiration_date=None
        )

    def test_valid_token_without_expiration(self):
        validated_token = self.authenticator.authenticate_credentials(
            self.token, self.project.url_id)[1]
        self.assertEqual(self.token, validated_token)

    def test_valid_token_with_expiration(self):
        self.token.delete()
        token = ProjectAccessToken.objects.create(
            project=self.project,
            created=timezone.now(),
            expiration_date=timezone.now() + datetime.timedelta(days=2)
        )
        validated_token = self.authenticator.authenticate_credentials(
            token, self.project.url_id)[1]
        self.assertEqual(token, validated_token)

    def test_invalid_token(self):
        token = 'rubbish'
        self.assertRaises(
            exceptions.AuthenticationFailed,
            self.authenticator.authenticate_credentials,
            token, self.project.url_id
        )

    def test_without_token(self):
        self.assertRaises(
            exceptions.AuthenticationFailed,
            self.authenticator.authenticate_credentials,
            None, self.project.url_id
        )

    def test_expired_token(self):
        self.token.delete()
        expired_token = ProjectAccessToken.objects.create(
            project=self.project,
            created=timezone.now(),
            expiration_date=datetime.datetime(2022, 2, 2, 22, 22).replace(tzinfo=datetime.timezone.utc)
        )
        self.assertRaises(
            exceptions.AuthenticationFailed,
            self.authenticator.authenticate_credentials,
            expired_token, self.project.url_id
        )



# Generated by Django 3.2.13 on 2024-10-25 08:40

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('ddm', '0048_alter_eventlogentry_description'),
    ]

    operations = [
        migrations.SeparateDatabaseAndState(
            state_operations = [
                migrations.DeleteModel(
                    name = 'ProjectAccessToken',
                ),
            ],
            database_operations = [
                migrations.AlterModelTable(
                    name = 'ProjectAccessToken',
                    table = 'ddm_auth_projectaccesstoken',
                ),
            ],
        )
    ]



# Generated by Django 4.2.16 on 2024-11-07 10:43

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('ddm_questionnaire', '0005_auto_20241026_1048'),
    ]

    operations = [
        migrations.AlterField(
            model_name='questionbase',
            name='polymorphic_ctype',
            field=models.ForeignKey(editable=False, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='polymorphic_%(app_label)s.%(class)s_set+', to='contenttypes.contenttype'),
        ),
        migrations.AlterField(
            model_name='questionbase',
            name='text',
            field=models.TextField(blank=True, default='', help_text='If a question is linked to a File Blueprint, data points from the donated data associated with the linked donation blueprint can be included in the question text. This data can be included as "{{ data }}" in the question text. It is possible to subset the data object (e.g., to include the last datapoint you can use {{ data.0 }} or include advanced rendering options included in the Django templating engine. For a more comprehensive overview and examples see the documentation. Additionally, information directly related to the participant can be included in the question text. This information can be referenced as "{{ participant }}".'),
            preserve_default=False,
        ),
    ]



from django.urls import path

from ddm.questionnaire import views


app_name = 'ddm_questionnaire'
urlpatterns = [
    path(r'', views.QuestionnaireOverview.as_view(), name='overview'),
    path(r'<slug:question_type>/create/', views.QuestionCreate.as_view(), name='create'),
    path(r'<slug:question_type>/<int:pk>/edit/', views.QuestionEdit.as_view(), name='edit'),
    path(r'<slug:question_type>/<int:pk>/delete/', views.QuestionDelete.as_view(), name='delete'),
    path(r'<slug:question_type>/<int:pk>/items/', views.ItemEdit.as_view(), name='items'),
    path(r'<slug:question_type>/<int:pk>/scale/', views.ScaleEdit.as_view(), name='scale'),
]



# Generated by Django 3.2.13 on 2024-10-25 15:49

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('ddm', '0053_auto_20241025_1451'),
    ]

    operations = [
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.RemoveField(
                    model_name='participant',
                    name='project',
                ),
            ],
            database_operations=[]
        ),
    ]



from django.contrib.auth import get_user_model
from django.test import TestCase
from django.utils import timezone

from ddm.logging.models import ExceptionLogEntry
from ddm.participation.models import Participant
from ddm.projects.models import DonationProject, ResearchProfile
from ddm.questionnaire.exceptions import QuestionValidationError
from ddm.questionnaire.models import (
    SingleChoiceQuestion, QuestionItem, MultiChoiceQuestion, MatrixQuestion,
    ScalePoint, SemanticDifferential, QuestionnaireResponse
)
from ddm.questionnaire.services import save_questionnaire_to_db

User = get_user_model()


class TestQuestionnaireServices(TestCase):
    @classmethod
    def setUpTestData(cls):
        user = User.objects.create_user(**{
            'username': 'owner', 'password': '123', 'email': 'owner@mail.com'
        })
        profile = ResearchProfile.objects.create(user=user)

        cls.project = DonationProject.objects.create(
            name='Base Project', slug='base', owner=profile)

        cls.participant = Participant.objects.create(
            project=cls.project, start_time=timezone.now())

        cls.question_config = {
            'project': cls.project,
            'name': 'Test Question',
            'page': 1,
            'index': 1,
            'variable_name': 'test_var',
            'text': 'Question Text',
        }

        cls.question = SingleChoiceQuestion.objects.create(**cls.question_config)
        cls.item_a = QuestionItem.objects.create(question=cls.question, index=1, value=1)
        cls.item_b =QuestionItem.objects.create(question=cls.question, index=2, value=8)

    def test_save_questionnaire_to_db_valid(self):
        valid_data = {
            f'{self.question.pk}': {
                'response': 1,
                'question': self.question.text,
                'items': [self.item_a.serialize_to_config(), self.item_b.serialize_to_config()],
            }
        }
        n_responses_before = QuestionnaireResponse.objects.count()
        n_logs_before = ExceptionLogEntry.objects.count()

        save_questionnaire_to_db(valid_data, self.project, self.participant)

        n_responses_after = QuestionnaireResponse.objects.count()
        n_logs_after = ExceptionLogEntry.objects.count()

        self.assertEqual(n_responses_before + 1, n_responses_after)
        self.assertEqual(n_logs_before, n_logs_after)

    def test_save_questionnaire_to_db_question_does_not_exist(self):
        invalid_data = {
            f'2': {
                'response': 1,
                'question': self.question.text,
                'items': [self.item_a.serialize_to_config(), self.item_b.serialize_to_config()],
            }
        }
        n_responses_before = QuestionnaireResponse.objects.count()
        n_logs_before = ExceptionLogEntry.objects.count()

        save_questionnaire_to_db(invalid_data, self.project, self.participant)

        n_responses_after = QuestionnaireResponse.objects.count()
        n_logs_after = ExceptionLogEntry.objects.count()

        self.assertEqual(n_responses_before + 1, n_responses_after)
        self.assertEqual(n_logs_before + 1, n_logs_after)

    def test_save_questionnaire_to_db_invalid_question_id(self):
        invalid_data = {
            f'invalid-question-id': {
                'response': 1,
                'question': self.question.text,
                'items': [self.item_a.serialize_to_config(), self.item_b.serialize_to_config()],
            }
        }
        n_responses_before = QuestionnaireResponse.objects.count()
        n_logs_before = ExceptionLogEntry.objects.count()

        save_questionnaire_to_db(invalid_data, self.project, self.participant)

        n_responses_after = QuestionnaireResponse.objects.count()
        n_logs_after = ExceptionLogEntry.objects.count()

        self.assertEqual(n_responses_before + 1, n_responses_after)
        self.assertEqual(n_logs_before + 1, n_logs_after)

    def test_save_questionnaire_to_db_invalid_response(self):
        valid_data = {
            f'{self.question.pk}': {
                'response': 2,
                'question': self.question.text,
                'items': [self.item_a.serialize_to_config(), self.item_b.serialize_to_config()],
            }
        }
        n_responses_before = QuestionnaireResponse.objects.count()
        n_logs_before = ExceptionLogEntry.objects.count()

        save_questionnaire_to_db(valid_data, self.project, self.participant)

        n_responses_after = QuestionnaireResponse.objects.count()
        n_logs_after = ExceptionLogEntry.objects.count()

        self.assertEqual(n_responses_before + 1, n_responses_after)
        self.assertEqual(n_logs_before + 1, n_logs_after)



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class DDMDataDonationConfig(AppConfig):
    default_auto_field = 'django.db.models.AutoField'
    name = 'ddm.datadonation'
    label = 'ddm_datadonation'
    verbose_name = _('DDM Data Donation')

    def ready(self):
        import ddm.datadonation.signals



from rest_framework import serializers

from ddm.datadonation.models import DataDonation
from ddm.encryption.serializers import SerializerDecryptionMixin


class DonationSerializer(SerializerDecryptionMixin, serializers.HyperlinkedModelSerializer):
    project = serializers.IntegerField(source='project.id')
    data = serializers.SerializerMethodField()
    participant = serializers.IntegerField(source='participant.id')

    class Meta:
        model = DataDonation
        fields = ['time_submitted', 'consent', 'status', 'data', 'project', 'participant']



# Generated by Django 3.2.13 on 2024-10-25 09:19

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('ddm', '0049_delete_projectaccesstoken'),
    ]

    operations = [
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.RemoveField(
                    model_name='exceptionlogentry',
                    name='blueprint',
                ),
                migrations.RemoveField(
                    model_name='exceptionlogentry',
                    name='participant',
                ),
                migrations.RemoveField(
                    model_name='exceptionlogentry',
                    name='project',
                ),
                migrations.DeleteModel(
                    name='EventLogEntry',
                ),
                migrations.DeleteModel(
                    name='ExceptionLogEntry',
                ),
            ],
            database_operations=[
                migrations.AlterModelTable(
                    name='EventLogEntry',
                    table='ddm_logging_eventlogentry'
                ),
                migrations.AlterModelTable(
                    name='ExceptionLogEntry',
                    table='ddm_logging_exceptionlogentry'
                ),
            ]
        )
    ]



from django.contrib.auth import get_user_model
from django.test import TestCase, override_settings, Client
from django.urls import reverse
from django.utils import timezone
from django.utils.timezone import localtime

from rest_framework.test import APIClient

from ddm.apis.serializers import (
    ParticipantSerializer, ResponseSerializer, ResponseSerializerWithSnapshot,
    ProjectSerializer
)
from ddm.datadonation.models import DataDonation, DonationBlueprint
from ddm.participation.models import Participant
from ddm.projects.models import DonationProject, ResearchProfile
from ddm.questionnaire.models import OpenQuestion, QuestionnaireResponse


User = get_user_model()


@override_settings(DDM_SETTINGS={'EMAIL_PERMISSION_CHECK':  r'.*(\.|@)mail\.com$', })
class TestAPIs(TestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        cls.base_creds = {
            'username': 'owner', 'password': '123', 'email': 'owner@mail.com'
        }
        base_user = User.objects.create_user(**cls.base_creds)
        base_profile = ResearchProfile.objects.create(
            user=base_user
        )

        cls.no_perm_creds = {
            'username': 'non-perm', 'password': '123', 'email': 'non-perm@mail.com'
        }
        no_perm_user = User.objects.create_user(**cls.no_perm_creds)
        no_perm_profile = ResearchProfile.objects.create(
            user=no_perm_user
        )

        cls.project_base = DonationProject.objects.create(
            name='Base Project',
            slug='base',
            owner=base_profile
        )
        cls.project_no_quest = DonationProject.objects.create(
            name='Project no Questionnaire',
            slug='no-quest',
            owner=base_profile
        )
        cls.project_alt = DonationProject.objects.create(
            name='Alt Project',
            slug='alternative',
            owner=no_perm_profile
        )
        cls.project_super_secret = DonationProject.objects.create(
            name='Secret Project',
            slug='super-secret',
            owner=base_profile,
            super_secret=True
        )

        cls.participant_a = Participant.objects.create(
            project=cls.project_base,
            start_time=timezone.now()
        )
        cls.participant_b = Participant.objects.create(
            project=cls.project_base,
            start_time=timezone.now()
        )
        cls.participant_c = Participant.objects.create(
            project=cls.project_alt,
            start_time=timezone.now()
        )

        cls.blueprint_a = DonationBlueprint.objects.create(
            project=cls.project_base,
            name='donation blueprint',
            expected_fields='"a", "b"',
            file_uploader=None
        )
        cls.blueprint_b = DonationBlueprint.objects.create(
            project=cls.project_base,
            name='donation blueprint',
            expected_fields='"a", "b"',
            file_uploader=None
        )

        cls.donation_pA_a = DataDonation.objects.create(
            project=cls.project_base,
            blueprint=cls.blueprint_a,
            participant=cls.participant_a,
            time_submitted=timezone.now(),
            consent=True,
            status='{}',
            data=['data1_pA_bpA', 'data2_pA_bpA']
        )
        cls.donation_pA_b = DataDonation.objects.create(
            project=cls.project_base,
            blueprint=cls.blueprint_b,
            participant=cls.participant_a,
            time_submitted=timezone.now(),
            consent=True,
            status='{}',
            data=['data1_pA_bpB', 'data2_pA_bpB']
        )

        cls.donation_pB_a = DataDonation.objects.create(
            project=cls.project_base,
            blueprint=cls.blueprint_a,
            participant=cls.participant_b,
            time_submitted=timezone.now(),
            consent=True,
            status='{}',
            data=['data1_pB_bpA', 'data2_pB_bpA']
        )
        cls.donation_pB_b = DataDonation.objects.create(
            project=cls.project_base,
            blueprint=cls.blueprint_b,
            participant=cls.participant_b,
            time_submitted=timezone.now(),
            consent=True,
            status='{}',
            data=['data1_pB_bpB', 'data2_pB_bpB']
        )

        q = OpenQuestion.objects.create(
            project=cls.project_base,
            name='open question',
            page=1,
            index=1,
            variable_name='open'
        )
        cls.q_response_a = QuestionnaireResponse.objects.create(
            project=cls.project_base,
            participant=cls.participant_a,
            time_submitted=timezone.now(),
            data=f'{{"{q.pk}": {{"response": "response_data", "question": "question text", "items": []}}}}'
        )

        cls.q_response_b = QuestionnaireResponse.objects.create(
            project=cls.project_base,
            participant=cls.participant_b,
            time_submitted=timezone.now(),
            data=f'{{"{q.pk}": {{"response": "response_data2", "question": "question text", "items": []}}}}'
        )

    def test_project_detail_api_get_with_valid_api_credentials(self):
        token = self.project_base.create_token()
        client = APIClient()
        client.credentials(HTTP_AUTHORIZATION='Token ' + token.key)

        expected_response = {
            'participants': [
                ParticipantSerializer(self.participant_a).data,
                ParticipantSerializer(self.participant_b).data,],
            'blueprints': [
                self.blueprint_a.get_config(),
                self.blueprint_b.get_config()
            ],
            'project': ProjectSerializer(self.project_base).data,
            'metadata': {
                'n_blueprints': 2,
                'n_participants': 2
            }
        }
        url = reverse('ddm_apis:project_overview', args=[self.project_base.url_id])
        response = client.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertDictEqual(response.data, expected_response)

    def test_donations_api_get_with_valid_api_credentials_specific_participant(self):
        token = self.project_base.create_token()
        client = APIClient()
        client.credentials(HTTP_AUTHORIZATION='Token ' + token.key)

        expected_response = {
            'blueprints': {
                f'{self.blueprint_a.pk}': {
                    'blueprint_name': f'{self.blueprint_b.name}',
                    'donations': [
                        {
                            'participant': f'{self.participant_a.external_id}',
                            'data': ['data1_pA_bpA', 'data2_pA_bpA'],
                            'time_submitted': f'{localtime(self.donation_pA_a.time_submitted).isoformat()}',
                            'status': f'{self.donation_pA_a.status}',
                            'consent': self.donation_pA_a.consent,
                        },
                    ],
                },
                f'{self.blueprint_b.pk}': {
                    'blueprint_name': f'{self.blueprint_b.name}',
                    'donations': [
                        {
                            'participant': f'{self.participant_a.external_id}',
                            'data': ['data1_pA_bpB', 'data2_pA_bpB'],
                            'time_submitted': f'{localtime(self.donation_pA_b.time_submitted).isoformat()}',
                            'status': f'{self.donation_pA_b.status}',
                            'consent': self.donation_pA_b.consent,
                        },
                    ],
                }
            },
            'metadata': {
                'n_blueprints': 2
            }
        }

        url = reverse(
            'ddm_apis:donations',
            args=[self.project_base.url_id]
        ) + f'?participants={self.participant_a.external_id},nonsense'
        response = client.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertDictEqual(response.data, expected_response)

    def test_donations_api_get_with_valid_api_credentials_specific_blueprint(self):
        token = self.project_base.create_token()
        client = APIClient()
        client.credentials(HTTP_AUTHORIZATION='Token ' + token.key)

        expected_response = {
            'blueprints': {
                f'{self.blueprint_a.pk}': {
                    'blueprint_name': f'{self.blueprint_b.name}',
                    'donations': [
                        {
                            'participant': f'{self.participant_a.external_id}',
                            'data': ['data1_pA_bpA', 'data2_pA_bpA'],
                            'time_submitted': f'{localtime(self.donation_pA_a.time_submitted).isoformat()}',
                            'status': f'{self.donation_pA_a.status}',
                            'consent': self.donation_pA_a.consent,
                        },
                        {
                            'participant': f'{self.participant_b.external_id}',
                            'data': ['data1_pB_bpA', 'data2_pB_bpA'],
                            'time_submitted': f'{localtime(self.donation_pB_a.time_submitted).isoformat()}',
                            'status': f'{self.donation_pB_a.status}',
                            'consent': self.donation_pB_a.consent,
                        }
                    ],
                },
            },
            'metadata': {
                'n_blueprints': 1
            }
        }
        url = reverse(
            'ddm_apis:donations',
            args=[self.project_base.url_id]
        )
        query_string = (
            f'?blueprints={self.blueprint_a.pk},123&'
            f'participants={self.participant_a.external_id},'
            f'{self.participant_b.external_id}'
        )
        response = client.get(url + query_string)
        self.assertEqual(response.status_code, 200)
        self.assertDictEqual(response.data, expected_response)

    def test_donations_api_get_with_valid_api_credentials_no_participants_404(self):
        token = self.project_base.create_token()
        client = APIClient()
        client.credentials(HTTP_AUTHORIZATION='Token ' + token.key)

        url = reverse('ddm_apis:donations', args=[self.project_base.url_id])
        response = client.get(url)
        self.assertEqual(response.status_code, 400)

    def test_donations_api_get_fails_with_invalid_api_credentials(self):
        token = self.project_alt.create_token()
        client = APIClient()
        client.credentials(HTTP_AUTHORIZATION='Token ' + token.key)

        url = reverse('ddm_apis:donations', args=[self.project_base.url_id])
        response = client.get(url)
        self.assertEqual(response.status_code, 401)

    def test_donations_api_fails_with_no_api_credentials_created(self):
        token = self.project_base.create_token()
        key = token.key
        token.delete()
        client = APIClient()
        client.credentials(HTTP_AUTHORIZATION='Token ' + key)
        url = reverse('ddm_apis:donations', args=[self.project_base.url_id])
        response = client.get(url)
        self.assertEqual(response.status_code, 401)

    def test_donations_api_returns_404_with_non_existing_project(self):
        token = self.project_base.create_token()
        key = token.key
        client = APIClient()
        client.credentials(HTTP_AUTHORIZATION='Token ' + key)
        url = reverse('ddm_apis:donations', args=['non-existing-id'])
        response = client.get(url)
        self.assertEqual(response.status_code, 404)

    def test_donations_api_returns_405_for_super_secret_project(self):
        token = self.project_super_secret.create_token()
        key = token.key
        client = APIClient()
        client.credentials(HTTP_AUTHORIZATION='Token ' + key)
        url = reverse('ddm_apis:donations', args=[self.project_super_secret.url_id])
        response = client.get(url)
        self.assertEqual(response.status_code, 405)

    def test_responses_api_get_with_valid_api_credentials(self):
        token = self.project_base.create_token()
        client = APIClient()
        client.credentials(HTTP_AUTHORIZATION='Token ' + token.key)

        expected_response = {
            'responses': [
                ResponseSerializer(self.q_response_a).data,
                ResponseSerializer(self.q_response_b).data,
            ],
            'metadata': {
                'n_responses': len(self.project_base.questionnaireresponse_set.all())
            }
        }
        url = reverse('ddm_apis:responses', args=[self.project_base.url_id])
        response = client.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertDictEqual(response.data, expected_response)

    def test_responses_api_get_with_valid_api_credentials_and_snapshot(self):
        token = self.project_base.create_token()
        client = APIClient()
        client.credentials(HTTP_AUTHORIZATION='Token ' + token.key)

        expected_response = {
            'responses': [
                ResponseSerializerWithSnapshot(self.q_response_a).data,
                ResponseSerializerWithSnapshot(self.q_response_b).data,
            ],
            'metadata': {
                'n_responses': len(self.project_base.questionnaireresponse_set.all())
            }
        }
        url = reverse('ddm_apis:responses', args=[self.project_base.url_id])
        query_string = f'?include_snapshot=true'
        response = client.get(url + query_string)
        self.assertEqual(response.status_code, 200)
        self.assertDictEqual(response.data, expected_response)

    def test_responses_api_get_csv_with_valid_api_credentials(self):
        token = self.project_base.create_token()
        client = APIClient()
        client.credentials(HTTP_AUTHORIZATION='Token ' + token.key)
        url = reverse(
            'ddm_apis:responses',
            args=[self.project_base.url_id]
        )
        query_string = f'?csv=true'
        response = client.get(url + query_string)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response['Content-Type'], 'text/csv')

    def test_responses_api_get_csv_with_no_questionnaire(self):
        token = self.project_no_quest.create_token()  # TODO: Change to no
        client = APIClient()
        client.credentials(HTTP_AUTHORIZATION='Token ' + token.key)
        url = reverse(
            'ddm_apis:responses',
            args=[self.project_no_quest.url_id]
        )
        query_string = f'?csv=true'
        response = client.get(url + query_string)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response['Content-Type'], 'text/csv')


    def test_participant_deletion_with_regular_login(self):
        self.client.login(**self.base_creds)

        new_participant = Participant.objects.create(
            project=self.project_base, start_time=timezone.now())
        external_id = new_participant.external_id

        url = reverse(
            'ddm_apis:participant_delete',
            args=[self.project_base.url_id, external_id]
        )
        response = self.client.delete(url)
        self.assertEqual(response.status_code, 200)
        self.assertIsNone(Participant.objects.filter(
            external_id=external_id).first())

        url = reverse(
            'ddm_apis:participant_delete',
            args=[self.project_base.url_id, 'nonsense_id']
        )
        response = self.client.delete(url)
        self.assertEqual(response.status_code, 404)

    def test_participant_deletion_fails_for_user_without_permission(self):
        self.client.login(**self.no_perm_creds)

        new_participant = Participant.objects.create(
            project=self.project_base, start_time=timezone.now())
        external_id = new_participant.external_id

        url = reverse(
            'ddm_apis:participant_delete',
            args=[self.project_base.url_id, external_id]
        )
        response = self.client.delete(url, follow=True)
        self.assertEqual(response.status_code, 403)
        self.assertIsNotNone(Participant.objects.filter(
            external_id=external_id).first())

    def test_participant_deletion_with_token(self):
        token = self.project_base.create_token()
        client = APIClient()
        client.credentials(HTTP_AUTHORIZATION='Token ' + token.key)

        new_participant = Participant.objects.create(
            project=self.project_base, start_time=timezone.now())
        external_id = new_participant.external_id

        url = reverse(
            'ddm_apis:participant_delete',
            args=[self.project_base.url_id, external_id]
        )
        response = client.delete(url)

        self.assertEqual(response.status_code, 200)
        self.assertIsNone(Participant.objects.filter(
            external_id=external_id).first())

        url = reverse(
            'ddm_apis:participant_delete',
            args=[self.project_base.url_id, 'some_bogus_id']
        )
        response = client.delete(url)

        self.assertEqual(response.status_code, 404)

    def test_participant_deletion_fails_with_invalid_token(self):
        token = self.project_alt.create_token()
        client = APIClient()
        client.credentials(HTTP_AUTHORIZATION='Token ' + token.key)

        new_participant = Participant.objects.create(
            project=self.project_base, start_time=timezone.now())
        external_id = new_participant.external_id

        url = reverse(
            'ddm_apis:participant_delete',
            args=[self.project_base.url_id, external_id]
        )
        response = client.delete(url, follow=True)

        self.assertEqual(response.status_code, 401)
        self.assertIsNotNone(Participant.objects.filter(
            external_id=external_id).first())

    def test_delete_project_data_with_regular_login_owner(self):
        self.client.login(**self.base_creds)
        url = reverse(
            'ddm_apis:project_data_delete',
            args=[self.project_base.url_id]
        )
        response = self.client.delete(url)
        self.assertEqual(response.status_code, 200)
        self.assertIsNotNone(response.data)

    def test_delete_project_data_fails_for_user_without_permission(self):
        self.client.login(**self.no_perm_creds)
        url = reverse(
            'ddm_apis:project_data_delete',
            args=[self.project_base.url_id]
        )
        response = self.client.delete(url, follow=True)
        self.assertEqual(response.status_code, 403)

    def test_delete_project_data_fails_with_valid_api_credentials(self):
        token = self.project_base.create_token()
        client = APIClient()
        client.credentials(HTTP_AUTHORIZATION='Token ' + token.key)
        url = reverse(
            'ddm_apis:project_data_delete',
            args=[self.project_base.url_id]
        )
        response = client.delete(url)
        self.assertEqual(response.status_code, 403)
        self.assertIsNotNone(response.data)

    def test_delete_project_data_fails_with_invalid_api_credentials(self):
        token = self.project_alt.create_token()
        client = APIClient()
        client.credentials(HTTP_AUTHORIZATION='Token ' + token.key)
        url = reverse(
            'ddm_apis:project_data_delete',
            args=[self.project_base.url_id]
        )
        response = client.delete(url)
        self.assertEqual(response.status_code, 403)

    def test_delete_project_data_with_no_api_credentials_created(self):
        token = self.project_base.create_token()
        key = token.key
        token.delete()
        client = APIClient()
        client.credentials(HTTP_AUTHORIZATION='Token ' + key)
        url = reverse(
            'ddm_apis:project_data_delete',
            args=[self.project_base.url_id]
        )
        response = client.delete(url)
        self.assertEqual(response.status_code, 403)

    def test_download_project_detail_view_valid_login(self):
        self.client.login(**self.base_creds)
        url = reverse(
            'ddm_apis:download_project_details',
            args=[self.project_base.url_id]
        )
        response = self.client.get(url, follow=True)
        self.assertEqual(response.status_code, 200)

    def test_download_project_detail_view_invalid_login(self):
        self.client.login(**self.no_perm_creds)
        url = reverse(
            'ddm_apis:download_project_details',
            args=[self.project_base.url_id]
        )
        response = self.client.get(url, follow=True)
        self.assertEqual(response.status_code, 403)

    def test_download_project_detail_view_not_logged_in(self):
        client = Client()
        url = reverse(
            'ddm_apis:download_project_details',
            args=[self.project_base.url_id]
        )
        response = client.get(url, follow=True)
        self.assertEqual(response.status_code, 403)



# Generated by Django 3.2.13 on 2024-10-25 15:49

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('ddm_datadonation', '0002_alter_datadonation_participant'),
        ('ddm_questionnaire', '0004_alter_questionnaireresponse_participant'),
        ('ddm_logging', '0003_alter_exceptionlogentry_participant'),
        ('ddm', '0054_remove_participant_project'),
    ]

    operations = [
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.DeleteModel(
                    name='Participant',
                ),
            ],
            database_operations=[
                migrations.AlterModelTable(
                    name='Participant',
                    table='ddm_participation_participant'
                ),
            ]
        )
    ]



from django import forms
from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError

from django_ckeditor_5.widgets import CKEditor5Widget

from ddm.projects.models import DonationProject

User = get_user_model()


class ProjectCreateForm(forms.ModelForm):
    project_password = forms.CharField(
        min_length=10,
        max_length=150,
        required=False,
        widget=forms.PasswordInput()
    )
    project_password_confirm = forms.CharField(
        min_length=10,
        max_length=150,
        required=False,
        widget=forms.PasswordInput()
    )

    class Meta:
        model = DonationProject
        fields = [
            'name',
            'slug',
            'super_secret',
            'owner',
            'contact_information',
            'data_protection_statement'
        ]
        widgets = {
            'owner': forms.HiddenInput(),
            'data_protection_statement': CKEditor5Widget(config_name='ddm_ckeditor'),
            'contact_information': CKEditor5Widget(config_name='ddm_ckeditor'),
        }

    field_order = [
        'name',
        'slug',
        'super_secret',
        'project_password',
        'project_password_confirm',
        'contact_information',
        'data_protection_statement'
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if 'initial' in kwargs:
            self.fields['owner'].initial = kwargs['initial']['owner']

    def clean(self):
        super_secret = self.data.get('super_secret', False)
        secret = self.data.get('project_password', None)
        secret_confirm = self.data.get('project_password_confirm', None)

        if super_secret and secret in ['', None]:
            raise ValidationError('Super secret project needs a project password.')

        if super_secret and (secret != secret_confirm):
            raise ValidationError('The two supplied project passwords do not match".')

        super().clean()

    def save(self, commit=True):
        project = super().save(commit=False)
        if project.super_secret:
            project.secret_key = self.data['project_password']
        project.save()
        return project


class ProjectEditForm(forms.ModelForm):
    class Meta:
        model = DonationProject
        fields = [
            'name',
            'slug',
            'active',
            'contact_information',
            'data_protection_statement',
            'url_parameter_enabled',
            'expected_url_parameters',
            'redirect_enabled',
            'redirect_target',
            'img_header_left',
            'img_header_right'
        ]
        widgets = {
            'data_protection_statement': CKEditor5Widget(config_name='ddm_ckeditor'),
            'contact_information': CKEditor5Widget(config_name='ddm_ckeditor'),
        }

    def clean(self):
        cleaned_data = super().clean()
        url_parameter_enabled = cleaned_data.get('url_parameter_enabled', False)
        expected_url_parameters = cleaned_data.get('expected_url_parameters', '')

        if url_parameter_enabled and not expected_url_parameters:
            self.add_error('expected_url_parameters',
                           'URL parameter is enabled but no parameter is defined to be extracted.')

        redirect_enabled = cleaned_data.get('redirect_enabled', False)
        redirect_target = cleaned_data.get('redirect_target', '')
        if redirect_enabled and not redirect_target:
            self.add_error('redirect_target',
                           'Redirect is enabled but no redirect target is defined.')

        return cleaned_data


class BriefingEditForm(forms.ModelForm):
    class Meta:
        model = DonationProject
        fields = [
            'briefing_text',
            'briefing_consent_enabled',
            'briefing_consent_label_yes',
            'briefing_consent_label_no'
        ]
        widgets = {
            'briefing_text': CKEditor5Widget(config_name='ddm_ckeditor'),
        }

    def clean(self):
        cleaned_data = super().clean()

        consent_enabled = cleaned_data.get('briefing_consent_enabled', False)
        consent_label_yes = cleaned_data.get('briefing_consent_label_yes', '')
        consent_label_no = cleaned_data.get('briefing_consent_label_no', '')
        consent_label_error_msg = 'When briefing consent is enabled, a consent label must be provided.'

        if consent_enabled and not consent_label_yes:
            self.add_error('briefing_consent_label_yes', consent_label_error_msg)

        if consent_enabled and not consent_label_no:
            self.add_error('briefing_consent_label_no', consent_label_error_msg)
        return cleaned_data

class DebriefingEditForm(forms.ModelForm):
    class Meta:
        model = DonationProject
        fields = ['debriefing_text']
        widgets = {
            'debriefing_text': CKEditor5Widget(config_name='ddm_ckeditor_temp_func'),
        }



from django.urls import path, include

from ddm.datadonation import views


instruction_patterns = (
    [
        path(
            r'',
            views.InstructionOverview.as_view(),
            name='overview'
        ),
        path(
            r'create/',
            views.InstructionCreate.as_view(),
            name='create'
        ),
        path(
            r'<int:pk>/edit/',
            views.InstructionEdit.as_view(),
            name='edit'
        ),
        path(
            r'<int:pk>/delete/',
            views.InstructionDelete.as_view(),
            name='delete'
        ),
    ],
    'instructions'
)

blueprint_patterns = (
    [
        path(
            r'create/',
            views.BlueprintCreate.as_view(),
            name='create'
        ),
        path(
            r'<int:pk>/edit/',
            views.BlueprintEdit.as_view(),
            name='edit'
        ),
        path(
            r'<int:pk>/delete/',
            views.BlueprintDelete.as_view(),
            name='delete'
        ),
    ],
    'blueprints'
)

uploader_patterns = (
    [
        path(
            r'create/',
            views.FileUploaderCreate.as_view(),
            name='create'
        ),
        path(
            r'<int:pk>/edit/',
            views.FileUploaderEdit.as_view(),
            name='edit'
        ),
        path(
            r'<int:pk>/delete/',
            views.FileUploaderDelete.as_view(),
            name='delete'
        ),
    ],
    'uploaders'
)

app_name = 'ddm_datadonation'
urlpatterns = [
    path(
        r'',
        views.DataDonationOverview.as_view(),
        name='overview'
    ),
    path(
        r'blueprint/',
        include(blueprint_patterns)
    ),
    path(
        r'file-uploader/',
        include(uploader_patterns)
    ),
    path(
        r'file-uploader/<int:file_uploader_pk>/instructions/',
        include(instruction_patterns)
    ),
    path(
        'download/<slug:participant_id>',
        views.DonationDownloadView.as_view(),
        name='download_donation'
    )
]



import time

from selenium import webdriver
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.by import By
from selenium.webdriver.firefox.options import Options
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.wait import WebDriverWait
from PIL import Image

BASE_URL = 'http://127.0.0.1:8000/'
PROJECT_ID = 'F1svPg1v'
PROJECT_SLUG = 'your-individual-url'
FILE_UPLOADER_ID = 27


def initialize_driver():
    firefox_options = Options()
    firefox_options.set_preference("intl.accept_languages", "en,en-US")  # Example: German
    firefox_options.set_preference("general.useragent.locale", "en-US")  # Set UI locale too

    driver = webdriver.Firefox(options=firefox_options)
    driver.set_window_size(3840, 2160)
    return driver

def login_to_ddm(driver, url):
    driver.get(url)

    username_field = driver.find_element(By.ID, 'id_username')
    username_field.send_keys('admin')

    password_field = driver.find_element(By.ID, 'id_password')
    password_field.send_keys('password')

    login_button = driver.find_element(By.ID, 'login-btn')
    login_button.click()
    return

def prepare_project(driver):
    url = f'projects/{PROJECT_ID}/data-donation/file-uploader/{FILE_UPLOADER_ID}/edit/'
    driver.get(BASE_URL + url)

    WebDriverWait(driver, 10).until(EC.element_to_be_clickable((By.CSS_SELECTOR, 'label[for="id_combined_consent"]')))
    checkbox_label = driver.find_element(By.CSS_SELECTOR, 'label[for="id_combined_consent"]')
    driver.maximize_window()
    time.sleep(1)

    checkbox = driver.find_element(By.ID, 'id_combined_consent')
    if checkbox.is_selected():
        checkbox_label.click()

    save_button = driver.find_element(By.CSS_SELECTOR, 'input[type="submit"][value="Save Uploader"]')
    time.sleep(1)
    save_button.click()
    return

def add_margins_to_sc(sc_path):
    # Load screenshot.
    img = Image.open(sc_path)
    width, height = img.size

    # Add margins on top and bottom.
    margin_top = 10
    margin_bottom = 20
    new_height = height + margin_top + margin_bottom
    new_img = Image.new("RGB", (width, new_height), color='white')
    new_img.paste(img, (0, margin_top))

    # Replace the original screenshot.
    new_img.save(sc_path)
    return

def highlight_edit(driver, element):
    driver.execute_script("""
        var parent = arguments[0];
        var child = parent.querySelector('.inline-edit');
        if (child) {
            child.style.background = 'yellow';
        }
    """, element)
    return element

def highlight_participation_overview_download(driver, element):
    driver.execute_script("""
        var parent = arguments[0];
        var child = parent.querySelector('#download-participation-overview');
        if (child) {
            child.style.background = 'yellow';
        }
    """, element)
    return element

def highlight_responses_download(driver, element):
    driver.execute_script("""
        var parent = arguments[0];
        var child = parent.querySelector('#download-questionnaire-responses');
        if (child) {
            child.style.background = 'yellow';
        }
    """, element)
    return element

def prepare_project_settings(driver):
    button = driver.find_element(By.ID, 'project-base-settings-accordion-btn')
    button.click()
    return

def access_donation_stage(driver):
    consent_button = driver.find_element(By.ID, 'consent_yes')
    consent_button.click()

    next_button = driver.find_element(By.CLASS_NAME, 'flow-btn')
    next_button.click()
    return

def upload_file(driver):
    file_path = r'C:\Users\nipfif\PyCharmProjects\DDM\docs\takeout-demo.zip'
    file_input = driver.find_element(By.NAME, 'ul-0')
    file_input.send_keys(file_path)

    # Wait until file has been processed
    WebDriverWait(driver, 10)
    return

def enable_all_in_one_consent(driver):
    url = f'projects/{PROJECT_ID}/data-donation/file-uploader/{FILE_UPLOADER_ID}/edit/'
    driver.get(BASE_URL + url)

    WebDriverWait(driver, 10).until(EC.element_to_be_clickable((By.CSS_SELECTOR, 'label[for="id_combined_consent"]')))
    checkbox_label = driver.find_element(By.CSS_SELECTOR, 'label[for="id_combined_consent"]')
    driver.maximize_window()
    time.sleep(1)

    checkbox = driver.find_element(By.ID, 'id_combined_consent')
    if not checkbox.is_selected():
        checkbox_label.click()

    save_button = driver.find_element(By.CSS_SELECTOR, 'input[type="submit"][value="Save Uploader"]')
    time.sleep(1)
    save_button.click()

    url = f'studies/{PROJECT_SLUG}/data-donation/'
    driver.get(BASE_URL + url)

    upload_file(driver)
    return

def agree_to_donate(driver, element=None):
    url = f'studies/{PROJECT_SLUG}/data-donation/'
    driver.get(BASE_URL + url)
    upload_file(driver)
    WebDriverWait(driver, 10).until(EC.element_to_be_clickable((By.CSS_SELECTOR, 'label[for="combined-donate-agree"]')))
    driver.execute_script("window.scrollTo(0, document.body.scrollHeight);")
    time.sleep(1)

    # Get the checkbox itself
    checkbox = driver.find_element(By.ID, 'combined-donate-agree')
    checkbox_label = driver.find_element(By.CSS_SELECTOR, 'label[for="combined-donate-agree"]')

    # Ensure the checkbox is visible
    if not checkbox.is_selected():
        actions = ActionChains(driver)
        actions.move_to_element(checkbox_label).perform()
        time.sleep(0.5)
        checkbox_label.click()

    WebDriverWait(driver, 10).until(EC.element_to_be_clickable((By.CLASS_NAME, 'flow-btn')))
    next_button = driver.find_element(By.CLASS_NAME, 'flow-btn')
    driver.maximize_window()
    driver.execute_script("arguments[0].scrollIntoView({ block: 'center' });", next_button)
    time.sleep(1)
    next_button.click()
    time.sleep(10)
    return element

def create_question_screenshots(driver, element):
    question_types = [
        'singlechoice',
        'multichoice',
        'matrix',
        'semanticdifferential',
        'textblock',
        'open'
    ]
    question_containers = driver.find_elements(By.CLASS_NAME, 'question-container')
    for idx, container in enumerate(question_containers, start=1):
        driver.execute_script("arguments[0].scrollIntoView({ block: 'center' });", container)
        sc_path = docs_module_paths['researchers'] + f'questionnaire_{question_types[idx-1]}.png'
        container.screenshot(sc_path)
    return element

def access_debriefing(driver):
    for i in range(20):
        WebDriverWait(driver, 10).until(EC.element_to_be_clickable((By.CLASS_NAME, 'flow-btn')))
        next_button = driver.find_element(By.CLASS_NAME, 'flow-btn')
        driver.maximize_window()
        driver.execute_script("arguments[0].scrollIntoView({ block: 'center' });", next_button)
        time.sleep(1)
        next_button.click()
        if 'debriefing' in driver.current_url:
            break
    return

def load_element_list():
    elements_to_capture = [
        {
            'module': 'researchers',
            'sc_name': 'project_list.png',
            'url': 'projects/',
            'element_id': 'ddm-main',
            'function_pre': None,
            'function_post': None,
        },
        {
            'module': 'researchers',
            'sc_name': 'project_create.png',
            'url': 'projects/create/',
            'element_id': 'ddm-main',
            'function_pre': None,
            'function_post': None,
        },
        {
            'module': 'researchers',
            'sc_name': 'project_hub.png',
            'url': f'projects/{PROJECT_ID}/',
            'element_id': 'ddm-main',
            'function_pre': None,
            'function_post': None,
        },
        {
            'module': 'researchers',
            'sc_name': 'project_hub_edit.png',
            'url': f'projects/{PROJECT_ID}/',
            'element_id': 'project-details',
            'function_pre': None,
            'function_post': highlight_edit,
        },
        {
            'module': 'researchers',
            'sc_name': 'project_settings.png',
            'url': f'projects/{PROJECT_ID}/edit/',
            'element_id': 'ddm-main',
            'function_pre': prepare_project_settings,
            'function_post': None,
        },
        {
            'module': 'researchers',
            'sc_name': 'project_configuration_section.png',
            'url': f'projects/{PROJECT_ID}/',
            'element_id': 'project-configuration',
            'function_pre': None,
            'function_post': None,
        },
        {
            'module': 'researchers',
            'sc_name': 'datadonation_admin_page.png',
            'url': f'projects/{PROJECT_ID}/data-donation',
            'element_id': 'ddm-main',
            'function_pre': None,
            'function_post': None,
        },
        # Screenshot Briefing Page
        {
            'module': 'researchers',
            'sc_name': 'briefing_page.png',
            'url': f'studies/{PROJECT_SLUG}/briefing/',
            'element_id': 'ddm-participation-main',
            'function_pre': None,
            'function_post': None,
        },
        # Screenshot Data Donation Page. Must be executed after briefing page screenshots.
        {
            'module': 'researchers',
            'sc_name': 'datadonation_page.png',
            'url': f'studies/{PROJECT_SLUG}/data-donation/',
            'element_id': 'ddm-participation-main',
            'function_pre': access_donation_stage,
            'function_post': None,
        },
        {
            'module': 'researchers',
            'sc_name': 'datadonation_after_upload_page.png',
            'url': f'studies/{PROJECT_SLUG}/data-donation/',
            'element_id': 'ddm-participation-main',
            'function_pre': upload_file,
            'function_post': None,
        },
        {
            'module': 'researchers',
            'sc_name': 'datadonation_after_upload_page_allinone.png',
            'url': f'studies/{PROJECT_SLUG}/data-donation/',
            'element_id': 'ddm-participation-main',
            'function_pre': enable_all_in_one_consent,
            'function_post': None,
        },
        # Screenshot Questionnaire. Must be executed after data donation screenshots.
        {
            'module': 'researchers',
            'sc_name': 'questionnaire_page.png',
            'url': f'studies/{PROJECT_SLUG}/questionnaire/',
            'element_id': 'ddm-participation-main',
            'function_pre': agree_to_donate,
            'function_post': create_question_screenshots,
        },
        # Screenshot Debriefing Page. Must be executed after questionnaire screenshots.
        {
            'module': 'researchers',
            'sc_name': 'debriefing_page.png',
            'url': f'studies/{PROJECT_SLUG}/questionnaire/',
            'element_id': 'ddm-participation-main',
            'function_pre': access_debriefing,
            'function_post': None,
        },
        {
            'module': 'researchers',
            'sc_name': 'data_center.png',
            'url': f'projects/{PROJECT_ID}/',
            'element_id': 'data-center',
            'function_pre': None,
            'function_post': None,
        },
        {
            'module': 'researchers',
            'sc_name': 'data_download_section.png',
            'url': f'projects/{PROJECT_ID}/',
            'element_id': 'data-download-section',
            'function_pre': None,
            'function_post': None,
        },
        {
            'module': 'researchers',
            'sc_name': 'download_participation_overview.png',
            'url': f'projects/{PROJECT_ID}/',
            'element_id': 'data-download-section',
            'function_pre': None,
            'function_post': highlight_participation_overview_download,
        },
        {
            'module': 'researchers',
            'sc_name': 'download_responses.png',
            'url': f'projects/{PROJECT_ID}/',
            'element_id': 'data-download-section',
            'function_pre': None,
            'function_post': highlight_responses_download,
        },
        {
            'module': 'researchers',
            'sc_name': 'project_log_section.png',
            'url': f'projects/{PROJECT_ID}/',
            'element_id': 'project-log-section',
            'function_pre': None,
            'function_post': None,
        },
        {
            'module': 'researchers',
            'sc_name': 'participation_statistics_section.png',
            'url': f'projects/{PROJECT_ID}/',
            'element_id': 'participation-statistics-section',
            'function_pre': None,
            'function_post': None,
        },
        {
            'module': 'researchers',
            'sc_name': 'danger_zone.png',
            'url': f'projects/{PROJECT_ID}/',
            'element_id': 'danger-zone',
            'function_pre': None,
            'function_post': None,
        },
    ]
    return elements_to_capture

def main():
    driver = initialize_driver()
    login_to_ddm(driver, BASE_URL)
    elements_to_capture = load_element_list()
    prepare_project(driver)
    for entry in elements_to_capture:
        entry_url = BASE_URL + entry['url']
        driver.get(entry_url)

        if entry['function_pre']:
            fun = entry['function_pre']
            fun(driver)
            time.sleep(2)

        element = driver.find_element(By.ID, entry['element_id'])

        if entry['function_post']:
            fun = entry['function_post']
            element = fun(driver, element)
            time.sleep(2)

        sc_path = docs_module_paths[entry['module']] + entry['sc_name']
        element.screenshot(sc_path)
        add_margins_to_sc(sc_path)

    driver.quit()

docs_module_paths = {
    'ROOT': 'docs/modules/ROOT/images/',
    'administrators': 'docs/modules/administrators/images/',
    'developers': 'docs/modules/developers/images/',
    'researchers': 'modules/researchers/images/',
}

if __name__ == '__main__':
    """ 
    Run this script to re-generate screenshots used in the documentation.
    
    Please note that you must start the local development server for this script
    to be able to run (i.e., ddm/test_project/> python manage.py runserver).
    
    Also note that:
     a. the screenshots will be taken for a project that already exists 
     in the development database. You have to refer to this project and its components 
     by adjusting the variables defined at the beginning of this document.
     
     b. there needs to be a file called 'takeout-demo.zip' in the docs folder 
     that contains a valid data donation for demonstration purposes.
    """
    main()



import random

from django.db import models
from django.forms import model_to_dict
from django.utils import timezone

from polymorphic.models import PolymorphicModel

from ddm.core.utils.user_content.template import render_user_content
from ddm.encryption.models import ModelWithEncryptedData
from ddm.datadonation.models import DataDonation
from ddm.questionnaire.exceptions import QuestionValidationError


class QuestionType(models.TextChoices):
    GENERIC = 'generic', 'Generic Question'
    SINGLE_CHOICE = 'single_choice', 'Single Choice Question'
    MULTI_CHOICE = 'multi_choice', 'Multi Choice Question'
    MATRIX = 'matrix', 'Matrix Question'
    SEMANTIC_DIFF = 'semantic_diff', 'Semantic Differential'
    OPEN = 'open', 'Open Question'
    TRANSITION = 'transition', 'Text Block'


class QuestionBase(PolymorphicModel):
    project = models.ForeignKey(
        'ddm_projects.DonationProject',
        on_delete=models.CASCADE
    )
    blueprint = models.ForeignKey(
        'ddm_datadonation.DonationBlueprint',
        on_delete=models.CASCADE,
        null=True,
        blank=True
    )

    DEFAULT_QUESTION_TYPE = QuestionType.GENERIC
    question_type = models.CharField(
        max_length=20,
        blank=False,
        choices=QuestionType.choices,
        default=DEFAULT_QUESTION_TYPE
    )

    name = models.CharField(max_length=255)
    page = models.PositiveIntegerField(default=1, verbose_name='Page')
    index = models.PositiveIntegerField(default=1, verbose_name='Index')

    variable_name = models.SlugField(
        max_length=50,
        null=False,
        verbose_name='Variable name for storing response',
        help_text='Will be used in the data export to identify responses to this question.'
    )

    text = models.TextField(
        blank=True,
        help_text=(
            'If a question is linked to a File Blueprint, data points from the donated data associated with the linked donation blueprint can be included in the question text. '
            'This data can be included as "{{ data }}" in the question text. '
            'It is possible to subset the data object (e.g., to include the last datapoint you can use {{ data.0 }} or include advanced '
            'rendering options included in the Django templating engine. For a more comprehensive overview and examples see the documentation. '
            'Additionally, information directly related to the participant can be included in the question text. '
            'This information can be referenced as "{{ participant }}".'
        )
    )
    required = models.BooleanField(default=False)

    class Meta:
        ordering = ['page', 'index']
        constraints = [
            models.UniqueConstraint(
                fields=['variable_name', 'project'],
                name='unique_varname_per_project'
            ),
        ]

    def __init__(self, *args, **kwargs):
        if not args:
            kwargs['question_type'] = self.DEFAULT_QUESTION_TYPE
        super().__init__(*args, **kwargs)

    def __str__(self):
        return self.name

    def is_general(self):
        return True if self.blueprint is None else False

    def get_config(self, participant_id, view):
        config = self.create_config()
        config = self.render_config_content(config, participant_id, view)
        return config

    def create_config(self):
        config = {
            'question': self.pk,
            'type': self.question_type,
            'page': self.page,
            'index': self.index,
            'text': self.text,
            'required': self.required,
            'items': [],
            'scale': [],
            'options': {}
        }
        return config

    def render_config_content(self, config, participant, view):
        """
        Renders references to donated data or participant data in question or
        item text configurations as html.
        """
        if self.is_general():
            donated_data = None
        else:
            data_donation = DataDonation.objects.get(
                participant=participant,
                blueprint=self.blueprint
            )
            donated_data = data_donation.get_decrypted_data(
                secret=self.project.secret_key, salt=self.project.get_salt())

        context = {}
        context.update(participant.get_context_data())
        context.update({'donated_data': donated_data})

        config['text'] = render_user_content(config['text'], context)
        for index, item in enumerate(config['items']):
            item['label'] = render_user_content(item['label'], context)
            item['label_alt'] = render_user_content(item['label_alt'], context)
        return config


class ItemMixin(models.Model):
    randomize_items = models.BooleanField(default=False)

    class Meta:
        abstract = True

    def create_config(self):
        config = super().create_config()
        config = self.add_item_config(config)
        return config

    def add_item_config(self, config):
        items = QuestionItem.objects.filter(question=self)
        for item in items:
            config['items'].append(item.serialize_to_config())

        if self.randomize_items:
            random.shuffle(config['items'])
        return config

    def validate_item_response(self, response):
        """
        Checks if all expected item ids are present in response.
        Expects the response to be a dictionary of the form:
        {'item id as string': <item response>, ...}

        Raises a KeyError if validation fails.
        """
        question_items = self.questionitem_set.all().values_list('id', flat=True)
        item_ids = [str(i) for i in list(question_items)]
        response_keys = [str(k) for k in response.keys()]
        if sorted(item_ids) == sorted(response_keys):
            return True

        if len(item_ids) > len(response_keys):
            missing_items = [i for i in item_ids if i not in response_keys]
            raise KeyError(f'Missing expected items: {missing_items}.')
        elif len(item_ids) < len(response_keys):
            unexpected_items = [k for k in response_keys if k not in item_ids]
            raise KeyError(
                f'Unexpected item ids for {self.DEFAULT_QUESTION_TYPE} '
                f'with ID {self.id}: {unexpected_items}.')
        else:
            raise KeyError(
                f'Received response items ({sorted(response_keys)}) do '
                f'not match the expected items ({sorted(item_ids)}).')


class ScaleMixin:
    def create_config(self):
        config = super().create_config()
        config = self.add_scale_config(config)
        return config

    def add_scale_config(self, config):
        scale_points = ScalePoint.objects.filter(question=self)
        for point in scale_points:
            config['scale'].append(point.serialize_to_config())
        return config

    def validate_scale_response(self, response):
        """
        Checks if all responses are possible values according to the scale
        definition.
        {'item id as string': <item response>, ...}

        Raises a KeyError if validation fails.
        """
        scale_points = self.scalepoint_set.all()
        valid_values = list(scale_points.values_list('value', flat=True))
        valid_values.append(-99)

        errors = []
        for item, value in response.items():
            if isinstance(value, float):
                raise QuestionValidationError(
                    question=self,
                    message=f'Response {value} is of type float.')

            try:
                value = int(value)
            except (ValueError, TypeError):
                errors.append(f'Response {value} for item {item} '
                              f'cannot be converted to int.')
                continue

            if int(value) not in valid_values:
                errors.append(f'Got invalid response "{item}: {value}"')

        return errors


class SingleChoiceQuestion(ItemMixin, QuestionBase):
    DEFAULT_QUESTION_TYPE = QuestionType.SINGLE_CHOICE

    def get_valid_values(self):
        """
        Valid response values include all related item values
        (QuestionItem.value) plus -99 which indicates that a question was not
        answered/skipped.
        """
        question_items = self.questionitem_set.all()
        valid_values = list(question_items.values_list('value', flat=True))
        valid_values.append(-99)
        return valid_values

    def validate_response(self, response):
        """
        Expects the response to be a single value as either an integer or a
        string that can be converted to an integer.
        """
        valid_values = self.get_valid_values()

        if isinstance(response, float):
            raise QuestionValidationError(
                question=self,
                message=f'{response} is of type float.')

        try:
            response = int(response)
        except (ValueError, TypeError):
            raise QuestionValidationError(
                question=self,
                message=f'{response} cannot be converted to int.')

        if response not in valid_values:
            raise QuestionValidationError(
                question=self,
                message=f'{response} invalid; valid values: {valid_values}.')
        return True


class MultiChoiceQuestion(ItemMixin, QuestionBase):
    DEFAULT_QUESTION_TYPE = QuestionType.MULTI_CHOICE

    def get_valid_values(self):
        """
        Valid response values include 0 (item not selected), 1 (item selected),
        and -99 which indicates that a question was not answered/skipped.
        """
        valid_values = [0, 1, -99]
        return valid_values

    def validate_response(self, response):
        """
        Expects the response to be a dictionary of the form:
        {'item id as string': <1/0/-99>, ...}
        """
        errors = []
        try:
            self.validate_item_response(response)
        except KeyError as e:
            errors.append(e)

        valid_values = self.get_valid_values()
        for item, value in response.items():
            if isinstance(value, float):
                errors.append(f'Got invalid response {value} for item {item}.')
                continue

            try:
                value = int(value)
            except (ValueError, TypeError):
                errors.append(f'Response {response} for item {item} '
                              f'cannot be converted to int.')
                continue

            if value not in valid_values:
                errors.append(f'Got invalid response {value} for item {item}.')

        if errors:
            raise QuestionValidationError(question=self, item_errors=errors)
        return True


class OpenQuestion(QuestionBase):
    DEFAULT_QUESTION_TYPE = QuestionType.OPEN

    class DisplayOptions(models.TextChoices):
        SMALL = 'small', 'Small'
        LARGE = 'large', 'Large'
    display = models.CharField(
        max_length=20,
        blank=False,
        choices=DisplayOptions.choices,
        default=DisplayOptions.SMALL,
        help_text='"Small" displays a one-line textfield, "Large" a multiline '
                  'textfield as input.'
    )

    class InputTypes(models.TextChoices):
        TEXT = 'text', 'Any text'
        NUMBER = 'numbers', 'Numbers only'
        EMAIL = 'email', 'Email address'
    input_type = models.CharField(
        max_length=20,
        blank=False,
        choices=InputTypes.choices,
        default=InputTypes.TEXT,
        verbose_name='Input type',
        help_text='Select the type of input allowed.'
    )

    max_input_length = models.IntegerField(
        verbose_name='Maximum input length',
        help_text=(
            "Participants' input cannot exceed this length. If empty, "
            "no input length restriction is enforced."
        ),
        blank=True,
        null=True,
        default=None
    )

    def create_config(self):
        config = super().create_config()
        config['options']['display'] = self.display
        config['options']['input_type'] = self.input_type
        config['options']['max_input_length'] = self.max_input_length
        return config

    def validate_response(self, response):
        """ Any value representable as a String is a valid response. """
        try:
            str(response)
        except TypeError:
            raise QuestionValidationError(question=self)
        return True


class MatrixQuestion(ScaleMixin, ItemMixin, QuestionBase):
    DEFAULT_QUESTION_TYPE = QuestionType.MATRIX

    def validate_response(self, response):
        """
        Expects the response to be a dictionary of the form:
        {'item id as string': <1/0/-99>, ...}
        """
        item_errors = []
        try:
            self.validate_item_response(response)
        except KeyError as e:
            item_errors.append(e)

        scale_errors = self.validate_scale_response(response)

        if scale_errors or item_errors:
            raise QuestionValidationError(
                question=self, scale_errors=scale_errors, item_errors=item_errors)
        return True


class SemanticDifferential(ScaleMixin, ItemMixin, QuestionBase):
    DEFAULT_QUESTION_TYPE = QuestionType.SEMANTIC_DIFF

    def validate_response(self, response):
        """
        Expects the response to be a dictionary of the form:
        {'item id as string': <1/0/-99>, ...}
        """
        item_errors = []
        try:
            self.validate_item_response(response)
        except KeyError as e:
            item_errors.append(e)

        scale_errors = self.validate_scale_response(response)

        if scale_errors or item_errors:
            raise QuestionValidationError(
                question=self, scale_errors=scale_errors, item_errors=item_errors)
        return True


class Transition(QuestionBase):
    DEFAULT_QUESTION_TYPE = QuestionType.TRANSITION

    def validate_response(self, response):
        """
        No validation necessary, as this is just a placeholder question type.
        """
        return True


class QuestionItem(models.Model):
    class Meta:
        ordering = ['index']
        constraints = [
            models.UniqueConstraint(
                fields=['index', 'question'],
                name='unique_item_index_per_question'
            ),
            models.UniqueConstraint(
                fields=['value', 'question'],
                name='unique_item_value_per_question'
            ),
        ]

    question = models.ForeignKey(
        'QuestionBase',
        on_delete=models.CASCADE
    )
    index = models.IntegerField()
    label = models.CharField(
        max_length=255,
        null=True,
        blank=True
    )
    label_alt = models.CharField(
        max_length=255,
        null=True,
        blank=True,
        verbose_name="Label Right"
    )
    value = models.IntegerField()
    randomize = models.BooleanField(default=False)

    def serialize_to_config(self):
        item_config = model_to_dict(self, exclude=['question'])
        return item_config


class ScalePoint(models.Model):
    class Meta:
        ordering = ['index']
        constraints = [
            models.UniqueConstraint(
                fields=['index', 'question'],
                name='unique_index_per_question'
            ),
        ]

    question = models.ForeignKey(
        'QuestionBase',
        on_delete=models.CASCADE
    )
    index = models.IntegerField()
    label = models.CharField(
        max_length=100,
        null=True,
        blank=True
    )
    value = models.IntegerField()
    add_border = models.BooleanField(default=False)

    def serialize_to_config(self):
        scale_config = model_to_dict(self, exclude=['question'])
        return scale_config


class QuestionnaireResponse(ModelWithEncryptedData):
    # Will only ever be deleted, when the project is deleted.
    project = models.ForeignKey('ddm_projects.DonationProject', on_delete=models.CASCADE)
    participant = models.ForeignKey('ddm_participation.Participant', on_delete=models.CASCADE)
    time_submitted = models.DateTimeField(default=timezone.now)
    data = models.BinaryField()



from django.db import migrations, connection
from django.contrib.contenttypes.models import ContentType


def update_polymorphic_ctype(apps, schema_editor):
    # Define a mapping from old model paths to new model paths
    content_type_mapping = {
        'ddm.matrixquestion': 'ddm_questionnaire.MatrixQuestion',
        'ddm.multichoicequestion': 'ddm_questionnaire.MultiChoiceQuestion',
        'ddm.openquestion': 'ddm_questionnaire.OpenQuestion',
        'ddm.semanticdifferential': 'ddm_questionnaire.SemanticDifferential',
        'ddm.singlechoicequestion': 'ddm_questionnaire.SingleChoiceQuestion',
        'ddm.transition': 'ddm_questionnaire.Transition',
    }

    cursor = connection.cursor()

    for old_model, new_model in content_type_mapping.items():
        old_app_label, old_model_name = old_model.split('.')
        new_app_label, new_model_name = new_model.split('.')

        # Get the old and new ContentTypes
        try:
            old_ctype = ContentType.objects.get(app_label=old_app_label, model=old_model_name.lower())
            new_ctype = ContentType.objects.get(app_label=new_app_label, model=new_model_name.lower())
        except ContentType.DoesNotExist:
            continue

        # Update the polymorphic_ctype_id directly using raw SQL
        sql = f"""
        UPDATE ddm_questionnaire_questionbase
        SET polymorphic_ctype_id = %s
        WHERE polymorphic_ctype_id = %s
        """
        cursor.execute(sql, [new_ctype.id, old_ctype.id])

class Migration(migrations.Migration):

    dependencies = [
        ('ddm_questionnaire', '0001_initial'),
        ('contenttypes', '0002_remove_content_type_name')
    ]

    operations = [
        migrations.RunPython(update_polymorphic_ctype),
    ]



from django.contrib.auth import get_user_model
from django.test import TestCase

from ddm.projects.forms import ProjectCreateForm, ProjectEditForm, BriefingEditForm
from ddm.projects.models import ResearchProfile, DonationProject

User = get_user_model()


class TestProjectCreateForm(TestCase):
    @classmethod
    def setUpTestData(cls):
        user_credentials = {
            'username': 'user', 'password': '123', 'email': 'base@mail.com'
        }
        user = User.objects.create_user(**user_credentials)
        cls.user_profile = ResearchProfile.objects.create(user=user)

    def test_form_raises_error_if_super_secret_without_secret(self):
        post_data = {
            'name': 'project',
            'slug': 'project',
            'super_secret': True,
            'secret': '',
            'owner': self.user_profile
        }
        form = ProjectCreateForm(post_data)
        self.assertFalse(form.is_valid())

    def test_form_raises_error_if_super_secret_not_provided_twice(self):
        post_data = {
            'name': 'project',
            'slug': 'project',
            'super_secret': True,
            'project_password': 'secret',
            'project_password_confirm': 'different secret',
            'owner': self.user_profile
        }
        form = ProjectCreateForm(post_data)
        self.assertFalse(form.is_valid())

    def test_valid_form(self):
        post_data = {
            'name': 'project',
            'slug': 'project',
            'super_secret': False,
            'project_password': '',
            'project_password_confirm': '',
            'owner': self.user_profile,
            'contact_information': 'Some info.',
            'data_protection_statement': 'Some statement.'
        }
        form = ProjectCreateForm(post_data)
        self.assertTrue(form.is_valid())
        n_projects_before = DonationProject.objects.count()
        form.save()
        n_projects_after = DonationProject.objects.count()
        self.assertEqual(n_projects_before + 1, n_projects_after)


class TestProjectEditForm(TestCase):
    @classmethod
    def setUpTestData(cls):
        user_credentials = {
            'username': 'user', 'password': '123', 'email': 'base@mail.com'
        }
        user = User.objects.create_user(**user_credentials)
        cls.user_profile = ResearchProfile.objects.create(user=user)

        cls.project = DonationProject.objects.create(
            name='Base Project', slug='base', owner=cls.user_profile)

    def test_form_valid(self):
        post_data = {
            'project': self.project.id,
            'name': 'project',
            'slug': 'project',
            'contact_information': 'Some info.',
            'data_protection_statement': 'Some statement.',
            'url_parameter_enabled': False,
            'expected_url_parameters': '',
            'redirect_enabled': False,
            'redirect_target': '',
        }
        form = ProjectEditForm(post_data)
        self.assertTrue(form.is_valid())

    def test_form_invalid_if_url_parameter_misspecified(self):
        post_data = {
            'project': self.project.id,
            'name': 'project',
            'slug': 'project',
            'contact_information': 'Some info.',
            'data_protection_statement': 'Some statement.',
            'url_parameter_enabled': True,
            'expected_url_parameters': '',
            'redirect_enabled': False,
            'redirect_target': '',
        }
        form = ProjectEditForm(post_data)
        self.assertFalse(form.is_valid())

    def test_form_invalid_if_redirect_misspecified(self):
        post_data = {
            'project': self.project.id,
            'name': 'project',
            'slug': 'project',
            'contact_information': 'Some info.',
            'data_protection_statement': 'Some statement.',
            'url_parameter_enabled': False,
            'expected_url_parameters': '',
            'redirect_enabled': True,
            'redirect_target': '',
        }
        form = ProjectEditForm(post_data)
        self.assertFalse(form.is_valid())


class TestBriefingEditForm(TestCase):
    @classmethod
    def setUpTestData(cls):
        user_credentials = {
            'username': 'user', 'password': '123', 'email': 'base@mail.com'
        }
        user = User.objects.create_user(**user_credentials)
        cls.user_profile = ResearchProfile.objects.create(user=user)

        cls.project = DonationProject.objects.create(
            name='Base Project', slug='base', owner=cls.user_profile)

    def test_form_valid(self):
        post_data = {
            'briefing_text': 'Some text.',
            'briefing_consent_enabled': False,
            'briefing_consent_label_yes': '',
            'briefing_consent_label_no': None
        }
        form = BriefingEditForm(post_data)
        self.assertTrue(form.is_valid())

    def test_form_invalid_if_consent_labels_not_provided(self):
        post_data = {
            'briefing_text': 'Some text.',
            'briefing_consent_enabled': True,
            'briefing_consent_label_yes': '',
            'briefing_consent_label_no': None
        }
        form = BriefingEditForm(post_data)
        self.assertFalse(form.is_valid())
        self.assertIn('briefing_consent_label_yes', form.errors.keys())
        self.assertIn('briefing_consent_label_no', form.errors.keys())



from django import forms
from django.core.exceptions import ValidationError
from django.forms import inlineformset_factory, TextInput, Textarea

from django_ckeditor_5.widgets import CKEditor5Widget

from ddm.datadonation.models import DonationBlueprint, ProcessingRule, FileUploader, DonationInstruction


class BlueprintEditForm(forms.ModelForm):

    class Meta:
        model = DonationBlueprint
        fields = [
            'name',
            'description',
            'regex_path',
            'exp_file_format',
            'csv_delimiter',
            'file_uploader',
            'json_extraction_root',
            'expected_fields',
            'expected_fields_regex_matching'
        ]
        widgets = {
            'expected_fields': forms.Textarea(attrs={'rows': 1}),
            'regex_path': forms.Textarea(attrs={'rows': 1}),
            'description': forms.Textarea(attrs={'rows': 3}),
        }

    def clean(self):
        related_file_uploader = self.data.get('file_uploader', None)
        regex = self.data.get('regex_path', None)

        if related_file_uploader:
            file_uploader = FileUploader.objects.get(pk=related_file_uploader)
            if file_uploader.upload_type == FileUploader.UploadTypes.ZIP_FILE and regex in ['', None]:
                raise ValidationError(
                    'Donation Blueprints that belong to a ZIP file uploader must define '
                    'a regex pattern.'
                )
        super().clean()


class InstructionsForm(forms.ModelForm):
    class Meta:
        model = DonationInstruction
        fields = ['text', 'index']
        widgets = {
            'text': CKEditor5Widget(config_name='ddm_ckeditor'),
        }


class ProcessingRuleForm(forms.ModelForm):

    class Meta:
        model = ProcessingRule
        fields = [
            'execution_order',
            'name',
            'field',
            'regex_field',
            'comparison_operator',
            'comparison_value',
            'replacement_value'
        ]
        widgets = {
            'field': TextInput(),
            'comparison_value': Textarea(attrs={'cols': 60, 'rows': 1}),
            'replacement_value': Textarea(attrs={'cols': 60, 'rows': 1}),
        }


ProcessingRuleInlineFormset = inlineformset_factory(
    DonationBlueprint,
    ProcessingRule,
    form=ProcessingRuleForm,
    extra=0
)


class SecretInputForm(forms.Form):
    secret = forms.CharField(
        widget=forms.PasswordInput()
    )



from django.contrib.auth import get_user_model
from django.test import TransactionTestCase

from ddm.datadonation.models import DonationInstruction, FileUploader
from ddm.projects.models import DonationProject, ResearchProfile


User = get_user_model()


class TestSignals(TransactionTestCase):
    @classmethod
    def setUp(self):
        user = User.objects.create_user(**{
            'username': 'owner', 'password': '123', 'email': 'owner@mail.com'
        })
        profile = ResearchProfile.objects.create(user=user)

        self.project = DonationProject.objects.create(
            name='Base Project', slug='base', owner=profile)

        self.file_uploader_a = FileUploader.objects.create(
            project=self.project,
            name='basic file uploader',
            upload_type=FileUploader.UploadTypes.SINGLE_FILE,
            index=1
        )

        self.file_uploader_b = FileUploader.objects.create(
            project=self.project,
            name='basic file uploader B',
            upload_type=FileUploader.UploadTypes.SINGLE_FILE,
            index=2
        )

        self.file_uploader_c = FileUploader.objects.create(
            project=self.project,
            name='basic file uploader C',
            upload_type=FileUploader.UploadTypes.SINGLE_FILE,
            index=3
        )

    def test_post_delete_signal_for_donation_instruction(self):
        DonationInstruction.objects.create(file_uploader=self.file_uploader_a, index=1)
        instruction_b = DonationInstruction.objects.create(file_uploader=self.file_uploader_a, index=2)
        instruction_c_pk = DonationInstruction.objects.create(file_uploader=self.file_uploader_a, index=3).pk
        instruction_b.delete()
        instruction_c = DonationInstruction.objects.get(pk=instruction_c_pk)
        self.assertEqual(instruction_c.index, 2)

    def test_post_delete_signal_for_file_uploader(self):
        file_uploader_c_pk = FileUploader.objects.create(
            project=self.project,
            name='basic file uploader C',
            upload_type=FileUploader.UploadTypes.SINGLE_FILE,
            index=3
        ).pk
        self.file_uploader_b.delete()
        file_uploader_c = FileUploader.objects.get(pk=file_uploader_c_pk)
        self.assertEqual(file_uploader_c.index, 2)



from django.contrib.auth import get_user_model
from django.test import TestCase

from ddm.datadonation.forms import BlueprintEditForm
from ddm.datadonation.models import FileUploader, DonationBlueprint
from ddm.projects.models import DonationProject, ResearchProfile

User = get_user_model()


class TestBlueprintEditForm(TestCase):
    @classmethod
    def setUpTestData(cls):
        user = User.objects.create_user(**{
            'username': 'owner', 'password': '123', 'email': 'owner@mail.com'
        })
        profile = ResearchProfile.objects.create(user=user)

        project = DonationProject.objects.create(
            name='Base Project', slug='base', owner=profile)

        cls.file_uploader = FileUploader.objects.create(
            project=project,
            name='zip file uploader',
            upload_type=FileUploader.UploadTypes.ZIP_FILE
        )

    def test_valid(self):
        bp = DonationBlueprint.objects.create(
            project=self.file_uploader.project,
            name='valid blueprint',
            description='some description',
            expected_fields='"some field"',
            file_uploader=self.file_uploader,
            regex_path='/this/file.json'
        )
        data = {
            'name': bp.name,
            'description': bp.description,
            'regex_path': bp.regex_path,
            'exp_file_format': 'json',
            'csv_delimiter': bp.csv_delimiter,
            'file_uploader': bp.file_uploader.pk,
            'json_extraction_root': bp.json_extraction_root,
            'expected_fields': bp.expected_fields,
            'expected_fields_regex_matching': bp.expected_fields_regex_matching,
        }
        form = BlueprintEditForm(data=data)
        self.assertTrue(form.is_valid())

    def test_invalid(self):
        bp = DonationBlueprint.objects.create(
            project=self.file_uploader.project,
            name='valid blueprint',
            description='some description',
            expected_fields='"some field"',
            file_uploader=self.file_uploader,
            regex_path='/this/file.json'
        )
        data = {
            'name': bp.name,
            'description': bp.description,
            'regex_path': '',
            'exp_file_format': 'json',
            'csv_delimiter': bp.csv_delimiter,
            'file_uploader': bp.file_uploader.pk,
            'json_extraction_root': bp.json_extraction_root,
            'expected_fields': bp.expected_fields,
            'expected_fields_regex_matching': bp.expected_fields_regex_matching,
        }
        form = BlueprintEditForm(data=data)
        self.assertFalse(form.is_valid())



# Generated by Django 3.2.13 on 2024-10-25 12:51

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('ddm_logging', '0002_alter_exceptionlogentry_blueprint'),
        ('ddm_questionnaire', '0003_alter_questionbase_blueprint'),
        ('ddm', '0052_auto_20241025_1451'),
    ]

    operations = [
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.DeleteModel(
                    name='DonationBlueprint',
                ),
                migrations.DeleteModel(
                    name='DonationInstruction',
                ),
                migrations.DeleteModel(
                    name='FileUploader',
                ),
                migrations.DeleteModel(
                    name='ProcessingRule',
                ),
            ],
            database_operations=[
                migrations.AlterModelTable(
                    name='DonationBlueprint',
                    table='ddm_datadonation_donationblueprint'
                ),
                migrations.AlterModelTable(
                    name='DonationInstruction',
                    table='ddm_datadonation_donationinstruction'
                ),
                migrations.AlterModelTable(
                    name='FileUploader',
                    table='ddm_datadonation_fileuploader'
                ),
                migrations.AlterModelTable(
                    name='ProcessingRule',
                    table='ddm_datadonation_processingrule'
                ),
            ]
        )
    ]



import json

from django.views.decorators.debug import sensitive_variables
from rest_framework import serializers

from ddm.datadonation.models import DataDonation
from ddm.encryption.serializers import SerializerDecryptionMixin
from ddm.participation.models import Participant
from ddm.projects.models import DonationProject
from ddm.questionnaire.models import QuestionnaireResponse, QuestionBase, QuestionItem


class ProjectSerializer(serializers.ModelSerializer):
    date_created = serializers.DateTimeField(format='iso-8601')

    class Meta:
        model = DonationProject
        fields = [
            'url_id',
            'slug',
            'name',
            'date_created',
            'active',
            'contact_information',
            'data_protection_statement',
            'url_parameter_enabled',
            'expected_url_parameters'
        ]


class DataDonationSerializer(SerializerDecryptionMixin, serializers.ModelSerializer):
    data = serializers.SerializerMethodField()
    participant = serializers.CharField(source='participant.external_id')
    time_submitted = serializers.DateTimeField(format='iso-8601')

    class Meta:
        model = DataDonation
        fields = [
            'participant',
            'data',
            'time_submitted',
            'status',
            'consent',
        ]


class ResponseSerializer(SerializerDecryptionMixin, serializers.ModelSerializer):
    participant = serializers.CharField(source='participant.external_id')
    response_data = serializers.SerializerMethodField()
    time_submitted = serializers.DateTimeField(format='iso-8601')

    class Meta:
        model = QuestionnaireResponse
        fields = [
            'participant',
            'response_data',
            'time_submitted'
        ]

    @sensitive_variables()
    def get_response_data(self, obj):
        """
        Creates a dictionary that only holds 'variable_name: response' pairs.
        """
        data = super().get_data(obj)
        try:
            data = json.loads(data)
        except TypeError:
            data = data

        responses = dict()
        for question_id in data.keys():
            try:
                question = QuestionBase.objects.all().get(id=question_id)
                var_name = question.variable_name
            except QuestionBase.DoesNotExist:
                # Question has been deleted.
                continue

            if isinstance(data[question_id]['response'], dict):
                item_answers = data[question_id]['response']
                for item_id in item_answers.keys():
                    try:
                        item = QuestionItem.objects.all().get(id=int(item_id))
                    except QuestionItem.DoesNotExist:
                        # Item has been deleted.
                        continue
                    value = item.value
                    responses[f'{var_name}-{value}'] = item_answers[item_id]
                pass
            else:
                responses[var_name] = data[question_id]['response']
        return responses


class ResponseSerializerWithSnapshot(ResponseSerializer):
    questionnaire_snapshot = serializers.SerializerMethodField()

    class Meta:
        model = QuestionnaireResponse
        fields = [
            'participant',
            'time_submitted',
            'questionnaire_snapshot',
            'response_data'
        ]

    @sensitive_variables()
    def get_questionnaire_snapshot(self, obj):
        data = super().get_data(obj)
        try:
            return json.loads(data)
        except TypeError:
            return data


class ParticipantSerializer(serializers.ModelSerializer):
    class Meta:
        model = Participant
        fields = [
            'external_id',
            'start_time',
            'end_time',
            'completed',
            'extra_data',
            'current_step'
        ]



from django import forms


class TokenCreationForm(forms.Form):
    expiration_days = forms.IntegerField(
        initial=30,
        min_value=1,
        max_value=90,
        required=True
    )
    action = forms.CharField(
        max_length=20,
        initial='create',
        widget=forms.HiddenInput()
    )



import json
import zipfile
from json import JSONDecodeError

from django.conf import settings
from django.http import HttpResponseRedirect
from django.shortcuts import redirect, reverse
from django.utils import timezone
from django.utils.datastructures import MultiValueDictKeyError
from django.utils.decorators import method_decorator
from django.utils.safestring import SafeString
from django.views.generic import TemplateView
from django.views.generic.detail import DetailView
from django.views.decorators.cache import cache_page

from ddm.core.utils.user_content.template import render_user_content
from ddm.datadonation.models import DonationBlueprint, FileUploader
from ddm.logging.utils import log_server_exception
from ddm.participation.models import Participant
from ddm.projects.models import DonationProject
from ddm.questionnaire.services import save_questionnaire_to_db


def get_participation_session_id(project):
    """
    Return id under which project related information is stored in a
    participant's session.
    """
    return f'project-{project.pk}'


def create_participation_session(request, project):
    """ Creates a new participation session if none does yet exist. """
    session_id = get_participation_session_id(project)

    if not request.session.get(session_id):
        participant = Participant.objects.create(
            project=project,
            start_time=timezone.now()
        )
        request.session[session_id] = {'participant_id': participant.id}
        request.session.modified = True
    return


class ParticipationFlowBaseView(DetailView):
    """
    Base class for participation flow views that implements base get and post
    methods, redirects and session management.
    """
    model = DonationProject
    context_object_name = 'project'

    steps = [
        'ddm_participation:briefing',
        'ddm_participation:datadonation',
        'ddm_participation:questionnaire',
        'ddm_participation:debriefing'
    ]
    participant = None
    current_step = None
    step_name = None

    def setup(self, request, *args, **kwargs):
        """ Initialize attributes shared by all participation steps. """
        super().setup(request, *args, **kwargs)
        self._initialize_values(request)

    def get(self, request, *args, **kwargs):
        # Check if project is active.
        if not self.object.active:
            return redirect(
                'ddm_participation:project_inactive', slug=self.object.slug)

        # Redirect to previous step if necessary.
        if not self.steps[self.current_step] == self.step_name:
            return redirect(self.steps[self.current_step], slug=self.object.slug)

        # Render current view.
        context = self.get_context_data(object=self.object)
        self.extra_before_render(request)
        return self.render_to_response(context)

    def post(self, request, *arges, **kwargs):
        # Account for 'page back' action in browser
        if self.steps[self.current_step] == self.step_name:
            self.set_step_completed()
            return redirect(
                self.steps[self.current_step + 1], slug=self.object.slug)
        else:
            return redirect(
                self.steps[self.current_step], slug=self.object.slug)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update({
            'default_header_left': getattr(settings, 'DDM_DEFAULT_HEADER_IMG_LEFT', None),
            'default_header_right': getattr(settings, 'DDM_DEFAULT_HEADER_IMG_RIGHT', None),
        })
        return context

    def get_participant_from_session(self, request):
        """
        Gets participant from session. If participant has not yet been created,
        creates new participant and saves it to session.
        """
        session_id = get_participation_session_id(self.object)
        participant_id = request.session[session_id]['participant_id']
        try:
            participant = Participant.objects.get(pk=participant_id)
        except Participant.DoesNotExist:
            participant = Participant.objects.create(
                project=self.object,
                start_time=timezone.now()
            )
            request.session[session_id]['participant_id'] = participant.id
            request.session.modified = True
        return participant

    @staticmethod
    def get_current_step_from_participant(participant):
        """
        Gets current step from information stored in participant's session.
        """
        step = participant.current_step
        if step is None:
            current_step = 0
            participant.current_step = 0
            participant.save()
        else:
            current_step = step
        return current_step

    def set_step_completed(self):
        """
        Updates the last_completed_step attribute in current session.
        """
        self.participant.current_step += 1
        self.participant.save()
        return

    def extra_before_render(self, request):
        """
        Placeholder for stage specific code executed before rendering the
        response.
        """
        pass

    def _initialize_values(self, request):
        self.object = self.get_object()
        create_participation_session(request, self.object)
        self.participant = self.get_participant_from_session(request)
        self.current_step = self.get_current_step_from_participant(self.participant)
        return


def participation_redirect_view(request, slug):
    """
    Redirect user to briefing page if url does not contain a step indicator.
    """
    redirect_url = reverse('ddm_participation:briefing', args=[slug])
    query_string = request.META.get('QUERY_STRING', '')
    return redirect(f'{redirect_url}?{query_string}')


class BriefingView(ParticipationFlowBaseView):
    template_name = 'ddm_participation/briefing.html'
    step_name = 'ddm_participation:briefing'

    def post(self, request, *args, **kwargs):
        """
        Checks whether participant has provided briefing consent to continue with
        the study. If briefing consent is not given, redirects to end page.
        If briefing consent is not within the expected values, the briefing page
        is again returned with a form error.
        """
        # Check briefing consent if enabled for project.
        if self.get_object().briefing_consent_enabled:
            return self.check_consent(request, **kwargs)

        return super().post(request, **kwargs)

    def check_consent(self, request, **kwargs):
        """
        Checks whether post data contains information on briefing consent.
        Renders briefing view with error message if consent information is invalid.
        Renders debriefing view if no consent has been given.
        Renders next step if consent has been given.
        """
        # Check that answer has been provided and is valid.
        consent = request.POST.get('briefing_consent', None)
        if consent not in ['0', '1']:
            # Render briefing view again with error message.
            context = self.get_context_data(object=self.object)
            context.update({'briefing_error': True})
            return self.render_to_response(context)

        # Save consent to participant data.
        self.participant.extra_data['briefing_consent'] = consent
        self.participant.save()

        if consent == '0':
            # Redirect to debriefing page.
            self.participant.current_step = len(self.steps) - 1
            self.participant.save()
            return redirect(self.steps[-1], slug=self.object.slug)
        else:
            return super().post(request, **kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        if self.object.url_parameter_enabled:
            self.extract_url_parameter()
        context['participant'] = self.participant
        participant_info = self.participant.get_context_data()
        context['briefing'] = render_user_content(self.object.briefing_text, participant_info)
        return context

    def extract_url_parameter(self):
        """
        Extract URL parameters on first call of the view and save to
        participant.extra_data.
        """
        if not self.participant.extra_data['url_param']:
            for param in self.object.get_expected_url_parameters():
                self.participant.extra_data['url_param'][param] = self.request.GET.get(param, None)
            self.participant.save()
        return


@method_decorator(cache_page(0), name='dispatch')
class DataDonationView(ParticipationFlowBaseView):
    template_name = 'ddm_participation/data_donation.html'
    step_name = 'ddm_participation:datadonation'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['uploader_configs'] = SafeString(self.get_uploader_configs())
        context['project_url_id'] = self.object.url_id
        return context

    def get_uploader_configs(self):
        project_uploaders = FileUploader.objects.filter(project=self.object)
        uploader_configs = [fu.get_configs(self.participant.get_context_data()) for fu in project_uploaders]
        return json.dumps(uploader_configs)

    def post(self, request, *args, **kwargs):
        super().post(request, **kwargs)  # TODO: Check if this is obsolete
        self.process_uploads(request.FILES)
        redirect_url = reverse(self.steps[self.current_step + 1], kwargs={'slug': self.object.slug})
        return HttpResponseRedirect(redirect_url)

    def process_uploads(self, files):
        try:
            file = files['post_data']
        except (MultiValueDictKeyError, KeyError) as e:
            msg = ('Data Donation Processing Exception: Did not receive '
                   f'expected data file from client. {e}')
            log_server_exception(self.object, msg)
            return

        if not zipfile.is_zipfile(file):
            msg = ('Data Donation Processing Exception: Data file received '
                   'from client is not a zip file.')
            log_server_exception(self.object, msg)
            return

        # Check if zip file contains expected file.
        unzipped_file = zipfile.ZipFile(file, 'r')
        if 'ul_data.json' not in unzipped_file.namelist():
            msg = 'Data Donation Processing Exception: "ul_data.json" is not in namelist.'
            log_server_exception(self.object, msg)
            return

        # Process donation data.
        try:
            file_data = json.loads(unzipped_file.read('ul_data.json').decode('utf-8'))
        except UnicodeDecodeError:
            try:
                file_data = json.loads(unzipped_file.read('ul_data.json').decode('latin-1'))
            except ValueError:
                msg = 'Donated data could not be decoded - tried both utf-8 and latin-1 decoding.'
                log_server_exception(self.object, msg)
                return
        except JSONDecodeError:
            msg = 'JSON decode error in donated data.'
            log_server_exception(self.object, msg)
            return

        for upload in file_data.keys():
            blueprint_id = upload
            blueprint_data = file_data[upload]
            try:
                blueprint = DonationBlueprint.objects.get(pk=blueprint_id,
                                                          project=self.object)
            except DonationBlueprint.DoesNotExist:
                msg = ('Data Donation Processing Exception: Referenced '
                       f'blueprint with id={blueprint_id} does not exist for '
                       'this project.')
                log_server_exception(self.object, msg)
                return
            blueprint.process_donation(blueprint_data, self.participant)


class QuestionnaireView(ParticipationFlowBaseView):
    template_name = 'ddm_participation/questionnaire.html'
    step_name = 'ddm_participation:questionnaire'

    def setup(self, request, *args, **kwargs):
        """ Reset extra_scripts """
        super().setup(request, *args, **kwargs)
        self.extra_scripts = []

    def get(self, request, *args, **kwargs):
        """
        Skip questionnaire if no questions are defined and redirect to next step.
        Otherwise, render questionnaire.
        """
        # Check if project is active.
        if not self.object.active:
            return redirect(
                'ddm_participation:project_inactive', slug=self.object.slug)

        # Redirect to previous step if necessary.
        if not self.steps[self.current_step] == self.step_name:
            return redirect(self.steps[self.current_step], slug=self.object.slug)

        context = self.get_context_data(object=self.object)
        if not len(context['q_config']) > 2:
            self.set_step_completed()
            return redirect(self.steps[self.current_step + 1],
                            slug=self.object.slug)
        else:
            return self.render_to_response(context)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        question_config = self.object.get_questionnaire_config(self.participant, self)
        context['q_config'] = json.dumps(question_config)
        context['extra_scripts'] = set(self.extra_scripts)
        return context

    def post(self, request, *args, **kwargs):
        super().post(request, **kwargs)
        self.process_response(request.POST)
        return redirect(self.steps[self.current_step + 1], slug=self.object.slug)

    def process_response(self, response):
        try:
            post_data = json.loads(response['post_data'])
        except MultiValueDictKeyError:
            msg = ('Questionnaire Post Exception: POST did not contain '
                   'expected key "post_data".')
            log_server_exception(self.object, msg)
            return

        save_questionnaire_to_db(post_data, self.object, self.participant)
        return


class DebriefingView(ParticipationFlowBaseView):
    template_name = 'ddm_participation/debriefing.html'
    step_name = 'ddm_participation:debriefing'

    def get_context_data(self, **kwargs):
        """ Inject url parameters in redirect target. """
        context = super().get_context_data(**kwargs)

        template_context = self.participant.get_context_data()
        template_context.update({'project_id': self.object.url_id})
        context['debriefing'] = render_user_content(self.object.debriefing_text, template_context)

        if self.object.redirect_enabled:
            context['redirect_target'] = render_user_content(self.object.redirect_target, template_context)
        else:
            context['redirect_target'] = None
        return context

    def extra_before_render(self, request):
        """Set step to completed and update participant information."""
        if not self.participant.completed:
            self.participant.end_time = timezone.now()
            self.participant.completed = True
            self.participant.save()
        return


class ContinuationView(DetailView):
    """
    Enables the continuation of the study at a later stage.
    Retrieves the session for a participant id passed as a URL parameter (?p=)
    and redirects to the last initiated stage of the study by the given
    participant.
    """
    model = DonationProject
    context_object_name = 'project'
    template_name = 'ddm_participation/continuation_not_found.html'

    def get(self, request, *args, **kwargs):
        self.object = self.get_object()
        context = super().get_context_data(**kwargs)
        participant_id = request.GET.get('p', None)
        try:
            participant = Participant.objects.get(external_id=participant_id)
        except Participant.DoesNotExist:
            # render continuation failed view with option to start new
            return self.render_to_response(context)

        self.initialize_session(request, participant.pk)
        return redirect(ParticipationFlowBaseView.steps[0], slug=self.object.slug)

    def initialize_session(self, request, participant_id):
        request.session[f'project-{self.object.pk}'] = {
            'participant_id': participant_id
        }
        request.session.modified = True
        return


class ProjectInactiveView(TemplateView):
    template_name = 'ddm_participation/project_inactive.html'



from django.contrib import auth
from django.http import Http404
from django.test import TestCase, override_settings, RequestFactory
from django.urls import reverse
from django.contrib.auth import get_user_model
from django.views.generic import TemplateView

from ddm.auth.views import DDMAuthMixin
from ddm.projects.models import ResearchProfile, DonationProject

User = get_user_model()


@override_settings(DDM_SETTINGS={'EMAIL_PERMISSION_CHECK':  r'.*(\.|@)mail\.com$', })
class TestNoPermissionView(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.valid_creds = {
            'username': 'owner', 'password': '123', 'email': 'owner@mail.com'
        }
        valid_user = User.objects.create_user(**cls.valid_creds)
        ResearchProfile.objects.create(user=valid_user)

        cls.not_permitted_creds = {
            'username': 'no_per', 'password': '123', 'email': 'noperm@liam.com'
        }
        not_permitted_user = User.objects.create_user(**cls.not_permitted_creds)
        ResearchProfile.objects.create(user=not_permitted_user)

    def test_ddm_no_permission_view_200_if_not_permitted(self):
        self.client.login(**self.not_permitted_creds)
        response = self.client.get(reverse('ddm_auth:no_permission'), follow=True)
        self.assertEqual(response.status_code, 200)

    def test_ddm_no_permission_view_200_if_anonymous(self):
        response = self.client.get(reverse('ddm_auth:no_permission'), follow=True)
        self.assertEqual(response.status_code, 200)

    def test_ddm_no_permission_view_redirect_to_project_list_if_permitted(self):
        self.client.login(**self.valid_creds)
        response = self.client.get(reverse('ddm_auth:no_permission'), follow=True)
        self.assertRedirects(response, reverse('ddm_projects:list'))


class TestProjectTokenView(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.valid_creds = {
            'username': 'owner', 'password': '123', 'email': 'owner@mail.com'
        }
        cls.valid_user = User.objects.create_user(**cls.valid_creds)
        valid_profile = ResearchProfile.objects.create(user=cls.valid_user)
        cls.project = DonationProject.objects.create(
            name='test-project', slug='test', owner=valid_profile)

        cls.not_permitted_creds = {
            'username': 'no_per', 'password': '123', 'email': 'noperm@liam.com'
        }
        not_permitted_user = User.objects.create_user(**cls.not_permitted_creds)
        ResearchProfile.objects.create(user=not_permitted_user)

    def test_200_if_permitted(self):
        self.client.login(**self.valid_creds)
        response = self.client.get(
            reverse('ddm_auth:project_token',
                    args=[self.project.url_id]),
            follow=True)
        self.assertEqual(response.status_code, 200)

    def test_redirect_if_not_permitted(self):
        self.client.login(**self.not_permitted_creds)
        response = self.client.get(
            reverse('ddm_auth:project_token',
                    args=[self.project.url_id]),
            follow=True)
        self.assertRedirects(response, reverse('ddm_auth:no_permission'))

    def test_redirect_if_anonymous(self):
        self.client.logout()
        response = self.client.get(
            reverse('ddm_auth:project_token',
                    args=[self.project.url_id]),
            follow=True)
        self.assertRedirects(response, reverse('ddm_login'))


class TestView(DDMAuthMixin, TemplateView):
    template_name = 'ddm_auth/no_permission.html'


@override_settings(DDM_SETTINGS={'EMAIL_PERMISSION_CHECK':  r'.*(\.|@)mail\.com$', })
class TestDDMAuthMixin(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.valid_creds = {
            'username': 'owner', 'password': '123', 'email': 'owner@mail.com'
        }
        cls.valid_user = User.objects.create_user(**cls.valid_creds)
        valid_profile = ResearchProfile.objects.create(user=cls.valid_user)
        cls.project = DonationProject.objects.create(
            name='test-project', slug='test', owner=valid_profile)

        cls.not_permitted_creds = {
            'username': 'no_per', 'password': '123', 'email': 'noperm@liam.com'
        }
        cls.not_permitted_user = User.objects.create_user(**cls.not_permitted_creds)
        ResearchProfile.objects.create(user=cls.not_permitted_user)

        cls.wo_profile_creds = {
            'username': 'no_prof', 'password': '123', 'email': 'noprof@mail.com'
        }
        cls.user_wo_profile = User.objects.create_user(**cls.wo_profile_creds)

        cls.not_owner_creds = {
            'username': 'notowner', 'password': '123', 'email': 'notowner@mail.com'
        }
        cls.not_owner = User.objects.create_user(**cls.not_owner_creds)
        ResearchProfile.objects.create(user=cls.not_owner)

        cls.view = TestView.as_view()
        cls.factory = RequestFactory()

    def test_dispatch_anonymous_user(self):
        anonymous_user = auth.get_user(self.client)
        request = self.factory.get('/test-url/')
        request.user = anonymous_user
        response = self.view(request)
        self.assertEqual(response.status_code, 302)
        self.assertTrue(response.url.startswith(reverse('ddm_login')))

    def test_dispatch_non_permitted_user(self):
        request = self.factory.get('/test-url/')
        request.user = self.not_permitted_user
        response = self.view(request)
        self.assertEqual(response.status_code, 302)
        self.assertTrue(response.url.startswith(reverse('ddm_auth:no_permission')))

    def test_dispatch_user_without_profile(self):
        request = self.factory.get('/test-url/')
        request.user = self.user_wo_profile
        self.assertFalse(ResearchProfile.objects.filter(user=self.user_wo_profile).exists())
        with self.assertRaises(Http404):
            response = self.view(request)
        self.assertTrue(ResearchProfile.objects.filter(user=self.user_wo_profile).exists())

    def test_dispatch_not_owner_user(self):
        request = self.factory.get('/test-url/')
        request.user = self.not_owner
        with self.assertRaises(Http404):
            self.view(request, **{'project_url_id': self.project.url_id})

    def test_dispatch_valid_user(self):
        request = self.factory.get('/test-url/')
        request.user = self.valid_user
        response = self.view(request, **{'project_url_id': self.project.url_id})
        self.assertEqual(response.status_code, 200)



from django.core.validators import MinLengthValidator
from django.db import models

from ddm.auth.models import ProjectAccessToken
from ddm.core.utils.misc import create_asciidigits_id
from ddm.datadonation.models import DataDonation
from ddm.logging.models import ExceptionLogEntry, EventLogEntry


def get_extra_data_default():
    """ Return default value for Participant.extra_data. """
    return dict(url_param=dict())


class Participant(models.Model):
    project = models.ForeignKey('ddm_projects.DonationProject', on_delete=models.CASCADE)

    external_id = models.CharField(
        unique=True, null=False,
        max_length=24,
        validators=[MinLengthValidator(24)]
    )

    # Participation statistics.
    start_time = models.DateTimeField()
    end_time = models.DateTimeField(null=True)
    completed = models.BooleanField(default=False)
    current_step = models.IntegerField(blank=True, null=True)

    extra_data = models.JSONField(default=get_extra_data_default)

    def get_context_data(self):
        """
        Returns data that can be accessed when participant is passed to a
        template as a context variable.
        """
        context_data = {
            'participant_id': self.external_id,
            'url_parameter': self.extra_data['url_param'],
            'donation_info': self.get_donation_info(),
        }
        if 'briefing_consent' in self.extra_data:
            context_data['briefing_consent'] = self.extra_data['briefing_consent']
        return context_data

    def get_donation_info(self):
        donations = DataDonation.objects.filter(participant=self)
        donation_info = {
            'n_success': donations.filter(status='success').count(),
            'n_pending': donations.filter(status='pending').count(),
            'n_failed': donations.filter(status='failed').count(),
            'n_consent': donations.filter(status='success', consent=True).count(),
            'n_no_consent': donations.filter(status='success', consent=False).count(),
            'n_no_data_extracted': donations.filter(status='nothing extracted').count()
        }
        return donation_info

    def save(self, *args, **kwargs):
        if self.pk is None:
            new_external_id = create_asciidigits_id(24)
            while len(Participant.objects.filter(external_id=new_external_id)) != 0:
                new_external_id = create_asciidigits_id(24)
            self.external_id = new_external_id
        super().save(*args, **kwargs)



# Generated by Django 3.2.13 on 2024-10-25 12:51

import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('ddm', '0052_auto_20241025_1451'),
    ]

    operations = [
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.CreateModel(
                    name='DonationBlueprint',
                    fields=[
                        ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                        ('name', models.CharField(help_text='Name for this File Blueprint. Will be visible to participants, so pick an informative name (e.g., "Watch History").', max_length=250)),
                        ('description', models.TextField(help_text='A description of which kind of data will be extracted by this Blueprint (e.g., "The title of your watched videos will be collected together with the time when you watched it."). Will be visible to participants.', null=True)),
                        ('exp_file_format', models.CharField(choices=[('json', 'JSON file'), ('csv', 'CSV file')], default='json', max_length=10, verbose_name='Expected file format')),
                        ('json_extraction_root', models.CharField(blank=True, default='', help_text="Indicates on which level of the files' data structure information should be extractet. If you want to extract information contained on the first level (e.g., {'field to be extracted': value}, you can leave this field empty. If you want to extract data located on a higher level, then you would provide the path to the parent field of the data you want to extract (e.g., if your json file is structured like this {'friends': {'real_friends': [{'name to extract': name, 'date to extract': date}], 'fake friends': [{'name': name, 'date': date }]}} and you want to extract the names and dates of real_friends, you would set the extraction root to 'friends.real_friends'.", max_length=200, verbose_name='Extraction Root')),
                        ('csv_delimiter', models.CharField(blank=True, default='', max_length=10)),
                        ('expected_fields', models.TextField(help_text='Put the field names in double quotes (") and separate them with commas ("Field A", "Field B").', validators=[django.core.validators.RegexValidator('^((["][^"]+["]))(\\s*,\\s*((["][^"]+["])))*[,\\s]*$', message='Field must contain one or multiple comma separated strings. Strings must be enclosed in double quotes ("string").')])),
                        ('expected_fields_regex_matching', models.BooleanField(default=False, help_text='Select if you use regex expressions in the "Excpected fields".')),
                        ('regex_path', models.TextField(blank=True, help_text='The path where the file is expected to be located in the uploaded ZIP folder. You can use Regex to, e.g., add wildcard characters or to match files in different languages. Consult the documentation for some examples.', null=True, verbose_name='File path')),
                    ],
                ),
                migrations.CreateModel(
                    name='ProcessingRule',
                    fields=[
                        ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                        ('name', models.CharField(help_text='An informative name for this rule. Only used internally.', max_length=250)),
                        ('field', models.TextField(help_text='The field on which the rule will be applied (just as a string without quotes).If a field is mentioned in a rule, it will be kept in the data that are sent to the server.')),
                        ('regex_field', models.BooleanField(default=False, help_text='Select if you use a regex expression in the "Field" setting to match a variable.')),
                        ('execution_order', models.IntegerField(help_text='The order in which the extraction steps are executed.')),
                        ('comparison_operator', models.CharField(blank=True, choices=[('', 'Keep Field'), ('==', 'Equal (==)'), ('!=', 'Not Equal (!=)'), ('>', 'Greater than (>)'), ('<', 'Smaller than (<)'), ('>=', 'Greater than or equal (>=)'), ('<=', 'Smaller than or equal (<=)'), ('regex-delete-match', 'Delete match (regex)'), ('regex-replace-match', 'Replace match (regex)'), ('regex-delete-row', 'Delete row when match (regex)')], default=None, max_length=24, null=True, verbose_name='Extraction Operator')),
                        ('comparison_value', models.TextField(blank=True, help_text='The value against which the data contained in the indicated field will be compared according to the selected comparison logic.')),
                        ('replacement_value', models.TextField(blank=True, help_text='Only required for operation "Replace match (regex)".')),
                        ('blueprint', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='ddm_datadonation.donationblueprint')),
                    ],
                ),
                migrations.CreateModel(
                    name='FileUploader',
                    fields=[
                        ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                        ('name', models.CharField(max_length=250)),
                        ('index', models.PositiveIntegerField()),
                        ('upload_type', models.CharField(choices=[('zip file', 'Zip File'), ('single file', 'Single File')], default='single file', max_length=20, verbose_name='Upload type')),
                        ('combined_consent', models.BooleanField(default=False, help_text='If enabled, participants will be asked to consent to submit all uploaded data at once. Otherwise, participant will be asked to consent to the submission of the data separately for each blueprint.', verbose_name='All-in-one consent')),
                        ('project', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='ddm.donationproject')),
                    ],
                ),
                migrations.CreateModel(
                    name='DonationInstruction',
                    fields=[
                        ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                        ('text', models.TextField(blank=True, null=True)),
                        ('index', models.PositiveIntegerField(default=1, validators=[django.core.validators.MinValueValidator(1)])),
                        ('file_uploader', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='ddm_datadonation.fileuploader', verbose_name='Associated File Uploader')),
                    ],
                    options={
                        'ordering': ['index'],
                    },
                ),
                migrations.AddField(
                    model_name='donationblueprint',
                    name='file_uploader',
                    field=models.ForeignKey(blank=True, help_text='The File Uploader through which the related file will be uploaded.', null=True, on_delete=django.db.models.deletion.SET_NULL, to='ddm_datadonation.fileuploader', verbose_name='Associated File Uploader'),
                ),
                migrations.AddField(
                    model_name='donationblueprint',
                    name='project',
                    field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='ddm.donationproject'),
                ),
                migrations.CreateModel(
                    name='DataDonation',
                    fields=[
                        ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                        ('time_submitted', models.DateTimeField(default=django.utils.timezone.now)),
                        ('consent', models.BooleanField(default=False)),
                        ('status', models.JSONField()),
                        ('data', models.BinaryField()),
                        ('blueprint', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='ddm_datadonation.donationblueprint')),
                        ('participant', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='ddm.participant')),
                        ('project', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='ddm.donationproject')),
                    ],
                    options={
                        'abstract': False,
                    },
                ),
                migrations.AddConstraint(
                    model_name='donationinstruction',
                    constraint=models.UniqueConstraint(fields=('index', 'file_uploader'), name='unique_index_per_file_uploader'),
                ),
            ],
            # Tables already exists. See ddm/migrations/0052 and 0053.
            database_operations=[]
        )
    ]



# Generated by Django 4.2.16 on 2025-02-15 11:44

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ddm_questionnaire', '0007_alter_questionitem_label_alt'),
    ]

    operations = [
        migrations.AddField(
            model_name='openquestion',
            name='input_type',
            field=models.CharField(choices=[('text', 'Any text'), ('numbers', 'Numbers only'), ('email', 'Email address')], default='text', help_text='Select the type of input allowed.', max_length=20, verbose_name='Input type'),
        ),
        migrations.AddField(
            model_name='openquestion',
            name='max_input_length',
            field=models.IntegerField(blank=True, default=None, help_text="Participants' input cannot exceed this length. If empty, no input length restriction is enforced.", null=True, verbose_name='Maximum input length'),
        ),
        migrations.AlterField(
            model_name='openquestion',
            name='display',
            field=models.CharField(choices=[('small', 'Small'), ('large', 'Large')], default='small', help_text='"Small" displays a one-line textfield, "Large" a multiline textfield as input.', max_length=20),
        ),
    ]



from django.urls import path

from ddm.logging import views
from ddm.logging import apis


app_name = 'ddm_logging'
urlpatterns = [
    path(r'projects/<slug:project_url_id>/logs/', views.ProjectLogsView.as_view(), name='project_logs'),
    path(r'<slug:project_url_id>/exceptions/', apis.ExceptionAPI.as_view(), name='exceptions_api'),
]



from django.template import Context, Template
from django.template.engine import Engine
from django.utils.safestring import SafeString


class TestEngine(Engine):
    default_builtins = ['ddm.core.utils.user_content.limited_template_library']


def preprocess_user_content(content: str) -> str:
    """
    Helper function to convert numeric html codes to characters to ensure the functionality of the custom
    templating features.
    """
    if not content:
        return ''

    content = content.replace('&gt;', '>')
    content = content.replace('&lt;', '<')
    return content


def render_user_content(content: str, context: dict=None) -> SafeString:
    """
    Function to render user provided content (used, e.g., to render a project's question text or debriefing text).

    Provides users (researchers) the flexibility to use a limited set of template tags to include variables and
    conditionals in their content.
    """
    content = preprocess_user_content(content)
    template = Template(content, engine=TestEngine())
    rendered_content = template.render(Context(context))
    return rendered_content



import json

from django.core.exceptions import ValidationError
from django.core.validators import MinValueValidator, RegexValidator
from django.db import models
from django.urls import reverse
from django.utils import timezone

from ddm.auth.models import ProjectAccessToken
from ddm.core.utils.user_content.template import render_user_content
from ddm.encryption.models import ModelWithEncryptedData
from ddm.logging.models import ExceptionLogEntry, ExceptionRaisers, EventLogEntry


COMMA_SEPARATED_STRINGS_VALIDATOR = RegexValidator(
    r'^((["][^"]+["]))(\s*,\s*((["][^"]+["])))*[,\s]*$',
    message=(
        'Field must contain one or multiple comma separated strings. '
        'Strings must be enclosed in double quotes ("string").'
    )
)


class FileUploader(models.Model):
    name = models.CharField(max_length=250)
    project = models.ForeignKey('ddm_projects.DonationProject', on_delete=models.CASCADE)
    index = models.PositiveIntegerField()

    class UploadTypes(models.TextChoices):
        ZIP_FILE = 'zip file'
        SINGLE_FILE = 'single file'

    upload_type = models.CharField(
        max_length=20,
        choices=UploadTypes.choices,
        default=UploadTypes.SINGLE_FILE,
        verbose_name='Upload type',
    )

    combined_consent = models.BooleanField(
        default=False,
        verbose_name='All-in-one consent',
        help_text='If enabled, participants will be asked to consent to submit '
                  'all uploaded data at once. Otherwise, participant will be asked to '
                  'consent to the submission of the data separately for each blueprint.'
    )

    def __str__(self):
        return self.name

    def delete(self, *args, **kwargs):
        """ This model has a post_delete signal processor (see signals.py). """
        super().delete(*args, **kwargs)

    def get_configs(self, participant_data=None):
        blueprints = self.donationblueprint_set.all()
        instructions = self.donationinstruction_set.all()
        configs = {
            'upload_type': self.upload_type,
            'name': self.name,
            'combined_consent': self.combined_consent,
            'blueprints': [bp.get_config() for bp in blueprints],
            'instructions': [{
                'index': i.index,
                'text': i.render(participant_data),
            } for i in instructions]
        }
        return configs

    def save(self, *args, **kwargs):
        if self.index is None:
            uploader_indices = self.project.fileuploader_set.all().values_list('index', flat=True)
            self.index = 1 if not uploader_indices else max(uploader_indices) + 1
        super().save(*args, **kwargs)


class DonationBlueprint(models.Model):
    project = models.ForeignKey(
        'ddm_projects.DonationProject',
        on_delete=models.CASCADE
    )
    name = models.CharField(
        max_length=250,
        help_text=(
            'Name for this File Blueprint. Will be visible to participants, '
            'so pick an informative name (e.g., "Watch History").'
        )
    )
    description = models.TextField(
        null=True,
        help_text=(
            'A description of which kind of data will be extracted by this '
            'Blueprint (e.g., "The title of your watched videos will be '
            'collected together with the time when you watched it."). '
            'Will be visible to participants.'
        )
    )

    class FileFormats(models.TextChoices):
        JSON_FORMAT = 'json', 'JSON file'
        CSV_FORMAT = 'csv', 'CSV file'
        # HTML_FORMAT = 'html',
        # XLSX_FORMAT = 'xlsx',

    exp_file_format = models.CharField(
        max_length=10,
        choices=FileFormats.choices,
        default=FileFormats.JSON_FORMAT,
        verbose_name='Expected file format',
    )

    json_extraction_root = models.CharField(
        max_length=200,
        default='',
        blank=True,
        verbose_name='Extraction Root',
        help_text=(
            'Indicates on which level of the files\' data structure information should be extractet. '
            'If you want to extract information contained on the first level (e.g., {\'field to be extracted\': value}, '
            'you can leave this field empty. If you want to extract data located on a higher level, then you would '
            'provide the path to the parent field of the data you want to extract (e.g., if your json file is structured like this '
            '{\'friends\': {\'real_friends\': [{\'name to extract\': name, \'date to extract\': date}], \'fake friends\': [{\'name\': name, \'date\': date }]}} '
            'and you want to extract the names and dates of real_friends, you would set the extraction root to \'friends.real_friends\'.'
        )
    )

    csv_delimiter = models.CharField(
        max_length=10,
        default="",
        blank=True,
        help_text=(
            'This field allows you to specify the character that separates '
            'values in the expected CSV file (e.g., , ; or \\t).'
            ' If left empty, DDM will try to infer the delimiter from the '
            'file structure.'
        )
    )

    expected_fields = models.TextField(
        null=False,
        blank=False,
        validators=[COMMA_SEPARATED_STRINGS_VALIDATOR],
        help_text=(
            'Put the field names in double quotes (") and separate them '
            'with commas ("Field A", "Field B").'
        )
    )

    expected_fields_regex_matching = models.BooleanField(
        default=False,
        null=False,
        help_text='Select if you use regex expressions in the "Expected fields".'
    )

    file_uploader = models.ForeignKey(
        'FileUploader',
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        verbose_name='Associated File Uploader',
        help_text=(
            'The File Uploader through which the related file will be uploaded.'
        )
    )
    regex_path = models.TextField(
        null=True,
        blank=True,
        verbose_name='File path',
        help_text=(
            'The path where the file is expected to be located in the uploaded '
            'ZIP folder. You can use Regex to, e.g., add wildcard characters or '
            'to match files in different languages. Consult the documentation '
            'for some examples.'
        )
    )

    def __str__(self):
        return self.name

    def get_absolute_url(self):
        return reverse('ddm_datadonation:blueprints:edit', args=[str(self.project.url_id), str(self.id)])

    def get_slug(self):
        return 'blueprint'

    def get_config(self):
        config = {
            'id': self.pk,
            'name': self.name,
            'description': self.description,
            'format': self.exp_file_format,
            'json_extraction_root': self.json_extraction_root,
            'expected_fields': json.loads("[" + str(self.expected_fields) + "]"),
            'exp_fields_regex_matching': self.expected_fields_regex_matching,
            'fields_to_extract': self.get_fields_to_extract(),
            'regex_path': self.regex_path,
            'filter_rules': self.get_filter_rules(),
            'csv_delimiter': self.csv_delimiter
        }
        return config

    def get_associated_questions(self):
        return self.questionbase_set.all()

    def get_filter_rules(self):
        return [r.get_rule_config() for r in self.processingrule_set.all().order_by('execution_order')]

    def get_fields_to_extract(self):
        fields = set()
        for r in self.processingrule_set.all():
            if r.comparison_operator is None:
                fields.add(r.field)
        return list(fields)

    def process_donation(self, data, participant):
        if self.validate_donation(data):
            self.create_donation(data, participant)
        else:
            msg = ('Data Donation Processing Exception: Donation validation '
                   f'failed for blueprint {self.pk}')
            ExceptionLogEntry.objects.create(
                project=self.project,
                blueprint=self,
                raised_by=ExceptionRaisers.SERVER,
                message=msg
            )
        return

    def validate_donation(self, data):
        # Check if all expected fields are in response.
        response_fields = ['consent', 'extracted_data', 'status']
        if not all(k in data for k in response_fields):
            msg = ('Data Donation Processing Exception: Donation data for '
                   f'Donation Blueprint {self.pk} does not contain the '
                   f'expected information. Expected fields: {response_fields}; '
                   f'Present fields: {data.keys()}.')
            ExceptionLogEntry.objects.create(
                project=self.project,
                blueprint=self,
                raised_by=ExceptionRaisers.SERVER,
                message=msg
            )
            return False

        return True

    def create_donation(self, data, participant):
        DataDonation.objects.create(
            project=self.project,
            blueprint=self,
            participant=participant,
            consent=data['consent'],
            status=data['status'],
            data=data['extracted_data']
        )
        return


class ProcessingRule(models.Model):
    """
    A processing rule that defines how the data uploaded to VUE will be processed
    before being sent to the server.
    Generates a json configuration that is passed to the VUE frontend component
    'UploaderApp'.
    """
    blueprint = models.ForeignKey(
        'DonationBlueprint',
        null=False,
        blank=False,
        on_delete=models.CASCADE
    )

    name = models.CharField(
        max_length=250,
        help_text='An informative name for this rule. Only used internally.'
    )

    field = models.TextField(
        null=False,
        blank=False,
        help_text=(
            'The field on which the rule will be applied (just as a string without quotes).'
            'If a field is mentioned in a rule, it will be kept in the data that are sent to the server.'
        )
    )
    regex_field = models.BooleanField(
        default=False,
        null=False,
        help_text='Select if you use a regex expression in the "Field" setting to match a variable.'
    )
    execution_order = models.IntegerField(
        help_text='The order in which the extraction steps are executed.'
    )

    class ComparisonOperators(models.TextChoices):
        EMPTY = '', 'Keep Field'
        EQUAL = '==', 'Equal (==)'
        NOT_EQUAL = '!=', 'Not Equal (!=)'
        GREATER = '>', 'Greater than (>)'
        SMALLER = '<', 'Smaller than (<)'
        GREATER_OR_EQUAL = '>=', 'Greater than or equal (>=)'
        SMALLER_OR_EQUAL = '<=', 'Smaller than or equal (<=)'
        REGEX_DELETE_MATCH = 'regex-delete-match', 'Delete match (regex)'
        REGEX_REPLACE_MATCH = 'regex-replace-match', 'Replace match (regex)'
        REGEX_DELETE_ROW = 'regex-delete-row', 'Delete row when match (regex)'

    comparison_operator = models.CharField(
        max_length=24,
        blank=True,
        null=True,
        choices=ComparisonOperators.choices,
        default=None,
        verbose_name='Extraction Operator'
    )
    comparison_value = models.TextField(
        blank=True,
        help_text='The value against which the data contained in the indicated field will '
                  'be compared according to the selected comparison logic.'
    )
    replacement_value = models.TextField(
        blank=True,
        help_text='Only required for operation "Replace match (regex)".'
    )

    def get_rule_config(self):
        """
        Return a configuration dict for the processing rule:
        {
            'field': 'field_name',
            'comparison_operator': '==' | '!=' | '>' | '<' | '>=' | '<=' |
                                   'regex-delete-match' | ' regex-replace-match'
                                   'regex-delete-row' | None,
            'comparison_value': '123' | Regex-String | None,
            'replacement_value': String
        }
        """
        return {
            'field': self.field,
            'regex_field': self.regex_field,
            'comparison_operator': self.comparison_operator,
            'comparison_value': self.comparison_value,
            'replacement_value': self.replacement_value
        }


class DataDonation(ModelWithEncryptedData):
    project = models.ForeignKey(
        'ddm_projects.DonationProject',
        on_delete=models.CASCADE
    )
    blueprint = models.ForeignKey(
        'DonationBlueprint',
        null=True,
        on_delete=models.SET_NULL
    )
    participant = models.ForeignKey(
        'ddm_participation.Participant',
        on_delete=models.CASCADE
    )
    time_submitted = models.DateTimeField(default=timezone.now)
    consent = models.BooleanField(default=False)
    status = models.JSONField()
    data = models.BinaryField()


class DonationInstruction(models.Model):
    text = models.TextField()
    index = models.PositiveIntegerField(default=1, validators=[MinValueValidator(1)])

    file_uploader = models.ForeignKey(
        'FileUploader',
        null=True,
        blank=False,
        on_delete=models.CASCADE,
        verbose_name='Associated File Uploader',
    )

    class Meta:
        ordering = ['index']
        constraints = [
            models.UniqueConstraint(
                fields=['index', 'file_uploader'],
                name='unique_index_per_file_uploader'
            ),
        ]

    def clean(self):
        # Ensure that index of instruction page is not greater than set of existing instructions + 1.
        n_instructions = self.file_uploader.donationinstruction_set.all().count()
        if self.pk:
            if self.index > n_instructions:
                raise ValidationError(
                    f'Index must be in range 1 to {n_instructions}.'
                )
        else:
            if self.index > (n_instructions + 1):
                raise ValidationError(
                    f'Index must be in range 1 to {n_instructions + 1}.'
                )
        super().clean()

    def save(self, *args, **kwargs):
        if kwargs.pop('ignore_index_check', False):
            return super().save()

        initial_index = DonationInstruction.objects.get(pk=self.pk).index if self.pk else None
        index_taken = self.file_uploader.donationinstruction_set.filter(index=self.index).exclude(pk=self.pk).exists()
        if index_taken and (self.index != initial_index):

            # Account for unique constraint by doing a "proxy"-save to free index.
            target_index = self.index
            self.index = self.file_uploader.donationinstruction_set.count() + 5
            super().save()

            # Change indices of involved objects:
            queryset = self.file_uploader.donationinstruction_set.exclude(pk=self.pk)
            if initial_index is None:
                queryset = queryset.filter(index__gte=target_index).order_by('-index')
                for q in queryset:
                    q.index += 1
                for q in queryset:
                    q.save(ignore_index_check=True)
            elif target_index < initial_index:
                queryset = queryset.filter(index__gte=target_index, index__lt=initial_index).order_by('-index')
                for q in queryset:
                    q.index += 1
                for q in queryset:
                    q.save(ignore_index_check=True)
            elif target_index > initial_index:
                queryset = queryset.filter(index__gt=initial_index, index__lte=target_index).order_by('index')
                for q in queryset:
                    q.index -= 1
                for q in queryset:
                    q.save(ignore_index_check=True)

            # Revert "proxy"-save.
            self.index = target_index
            return super().save()

        return super().save()

    def delete(self, *args, **kwargs):
        """ This model has a post_delete signal processor (see signals.py). """
        super().delete(*args, **kwargs)

    def render(self, context=None):
        return render_user_content(self.text, context)



# Generated by Django 3.2.13 on 2024-10-25 10:09

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('ddm', '0050_auto_20241025_1119'),
    ]

    operations = [
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.RemoveField(
                    model_name='multichoicequestion',
                    name='questionbase_ptr',
                ),
                migrations.RemoveField(
                    model_name='openquestion',
                    name='questionbase_ptr',
                ),
                migrations.RemoveField(
                    model_name='questionbase',
                    name='blueprint',
                ),
                migrations.RemoveField(
                    model_name='questionbase',
                    name='polymorphic_ctype',
                ),
                migrations.RemoveField(
                    model_name='questionbase',
                    name='project',
                ),
                migrations.RemoveField(
                    model_name='questionitem',
                    name='question',
                ),
                migrations.RemoveField(
                    model_name='questionnaireresponse',
                    name='participant',
                ),
                migrations.RemoveField(
                    model_name='questionnaireresponse',
                    name='project',
                ),
                migrations.RemoveField(
                    model_name='scalepoint',
                    name='question',
                ),
                migrations.RemoveField(
                    model_name='semanticdifferential',
                    name='questionbase_ptr',
                ),
                migrations.RemoveField(
                    model_name='singlechoicequestion',
                    name='questionbase_ptr',
                ),
                migrations.RemoveField(
                    model_name='transition',
                    name='questionbase_ptr',
                ),
                migrations.DeleteModel(
                    name='MatrixQuestion',
                ),
                migrations.DeleteModel(
                    name='MultiChoiceQuestion',
                ),
                migrations.DeleteModel(
                    name='OpenQuestion',
                ),
                migrations.DeleteModel(
                    name='QuestionBase',
                ),
                migrations.DeleteModel(
                    name='QuestionItem',
                ),
                migrations.DeleteModel(
                    name='QuestionnaireResponse',
                ),
                migrations.DeleteModel(
                    name='ScalePoint',
                ),
                migrations.DeleteModel(
                    name='SemanticDifferential',
                ),
                migrations.DeleteModel(
                    name='SingleChoiceQuestion',
                ),
                migrations.DeleteModel(
                    name='Transition',
                ),
            ],
            database_operations=[
                migrations.AlterModelTable(
                    name='MatrixQuestion',
                    table='ddm_questionnaire_matrixquestion'
                ),
                migrations.AlterModelTable(
                    name='MultiChoiceQuestion',
                    table='ddm_questionnaire_multichoicequestion'
                ),
                migrations.AlterModelTable(
                    name='OpenQuestion',
                    table='ddm_questionnaire_openquestion'
                ),
                migrations.AlterModelTable(
                    name='QuestionBase',
                    table='ddm_questionnaire_questionbase'
                ),
                migrations.AlterModelTable(
                    name='QuestionItem',
                    table='ddm_questionnaire_questionitem'
                ),
                migrations.AlterModelTable(
                    name='QuestionnaireResponse',
                    table='ddm_questionnaire_questionnaireresponse'
                ),
                migrations.AlterModelTable(
                    name='ScalePoint',
                    table='ddm_questionnaire_scalepoint'
                ),
                migrations.AlterModelTable(
                    name='SemanticDifferential',
                    table='ddm_questionnaire_semanticdifferential'
                ),
                migrations.AlterModelTable(
                    name='SingleChoiceQuestion',
                    table='ddm_questionnaire_singlechoicequestion'
                ),
                migrations.AlterModelTable(
                    name='Transition',
                    table='ddm_questionnaire_transition'
                ),
            ]
        )
    ]



from rest_framework.views import APIView
from rest_framework.response import Response
from rest_framework import permissions

from ddm.datadonation.models import DonationBlueprint
from ddm.logging.models import ExceptionLogEntry
from ddm.projects.models import DonationProject
from ddm.participation.models import Participant


class ExceptionAPI(APIView):

    permission_classes = [permissions.AllowAny]

    def post(self, request, format=None, *args, **kwargs):
        """
        Log exception messages received from client.
        """
        project_url_id = self.kwargs['project_url_id']
        project = DonationProject.objects.get(url_id=project_url_id)

        try:
            participant_id = request.session[f'project-{project.pk}']['participant_id']
            participant = Participant.objects.get(pk=participant_id)
        except KeyError:
            participant = None

        blueprint_id = request.data.get('blueprint')
        if blueprint_id:
            blueprint = DonationBlueprint.objects.get(pk=blueprint_id)
        else:
            blueprint = None

        ExceptionLogEntry.objects.create(
            project=project,
            participant=participant,
            exception_type=request.data.get('status_code'),
            message=request.data['message'],
            raised_by=request.data.get('raised_by'),
            blueprint=blueprint
        )

        return Response(None, status=201)



# Generated by Django 4.2.16 on 2024-12-05 13:29

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ddm_questionnaire', '0006_alter_questionbase_polymorphic_ctype_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='questionitem',
            name='label_alt',
            field=models.CharField(blank=True, max_length=255, null=True, verbose_name='Label Right'),
        ),
    ]



# Generated by Django 3.2.13 on 2024-10-25 15:49

import ddm.participation.models
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('ddm', '0054_remove_participant_project'),
    ]

    operations = [
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.CreateModel(
                    name='Participant',
                    fields=[
                        ('id',
                         models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                        ('external_id', models.CharField(max_length=24, unique=True,
                                                         validators=[django.core.validators.MinLengthValidator(24)])),
                        ('start_time', models.DateTimeField()),
                        ('end_time', models.DateTimeField(null=True)),
                        ('completed', models.BooleanField(default=False)),
                        ('current_step', models.IntegerField(blank=True, null=True)),
                        ('extra_data', models.JSONField(default=ddm.participation.models.get_extra_data_default)),
                        ('project',
                         models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='ddm.donationproject')),
                    ],
                ),
            ],
            # Table already exists. See ddm/migrations/0054, 0055.
            database_operations=[]
        )
    ]



# Generated by Django 3.2.13 on 2024-10-26 08:48

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('ddm_participation', '0002_alter_participant_project'),
        ('ddm_questionnaire', '0005_auto_20241026_1048'),
        ('ddm_datadonation', '0003_auto_20241026_1048'),
        ('ddm_logging', '0004_auto_20241026_1048'),
        ('ddm_auth', '0002_alter_projectaccesstoken_project'),
        ('ddm', '0055_delete_participant'),
    ]

    operations = [
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.RemoveField(
                    model_name='researchprofile',
                    name='user',
                ),
                migrations.DeleteModel(
                    name='DonationProject',
                ),
                migrations.DeleteModel(
                    name='ResearchProfile',
                ),
            ],
            database_operations=[
                migrations.AlterModelTable(
                    name='DonationProject',
                    table='ddm_projects_donationproject'
                ),
                migrations.AlterModelTable(
                    name='ResearchProfile',
                    table='ddm_projects_researchprofile'
                ),
            ]
        )
    ]
